# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from nova.tests.fixtures import libvirt as fakelibvirt
from nova.tests.functional.libvirt import test_vgpu


class VGPUTestsListDevices(test_vgpu.VGPUTestBase):
    """Regression test for bug 2098892.

    Test that nodeDeviceLookupByName() is called with valid types to prevent:

    File "/usr/lib64/python3.9/site-packages/libvirt.py", line 5201, in
      nodeDeviceLookupByName ret =
      libvirtmod.virNodeDeviceLookupByName(self._o, name)
      TypeError: virNodeDeviceLookupByName() argument 2 must be str or None,
      not Proxy

    in the future. This test relies on the LibvirtFixture checking for the
    correct types in its nodeDeviceLookupByName() method and raising TypeError
    if they are invalid.

    We don't test this by importing the libvirt module because the libvirt
    module is forbidden to be imported into our test environment. It is
    excluded from test-requirements.txt and we also use the
    ImportModulePoisonFixture in nova/test.py to prevent use of modules such as
    libvirt.
    """

    def setUp(self):
        super().setUp()

        # Start compute supporting only nvidia-11
        self.flags(
            enabled_mdev_types=fakelibvirt.NVIDIA_11_VGPU_TYPE,
            group='devices')

        self.start_compute_with_vgpu('host1')

    def test_update_available_resource(self):
        # We only want to verify no errors were logged by
        # update_available_resource (logging under the 'except Exception:').
        self._run_periodics(raise_on_error=True)



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime

from oslo_config import cfg
from oslo_utils.fixture import uuidsentinel as uuids

import nova
from nova.tests.functional.libvirt import base

CONF = cfg.CONF


class LibvirtVifModelTest(base.ServersTestBase):
    ADMIN_API = True

    def setUp(self):
        CONF.set_default("image_metadata_prefilter", True, group='scheduler')
        super().setUp()

        self.glance.create(
            None,
            {
                'id': uuids.image_vif_model_igb,
                'name': 'image-with-igb',
                'created_at': datetime.datetime(2011, 1, 1, 1, 2, 3),
                'updated_at': datetime.datetime(2011, 1, 1, 1, 2, 3),
                'deleted_at': None,
                'deleted': False,
                'status': 'active',
                'is_public': False,
                'container_format': 'bare',
                'disk_format': 'qcow2',
                'size': '74185822',
                'min_ram': 0,
                'min_disk': 0,
                'protected': False,
                'visibility': 'public',
                'tags': [],
                'properties': {
                    'hw_vif_model': 'igb',
                },
            }
        )

    def test_boot_with_vif_model_igb(self):
        orig_create = nova.virt.libvirt.guest.Guest.create
        self.xml = ""

        def fake_create(cls, xml, host):
            self.xml = xml
            return orig_create(xml, host)

        self.stub_out('nova.virt.libvirt.guest.Guest.create', fake_create)

        self.start_compute(
            hostname='compute1',
            libvirt_version=9003000,
            qemu_version=8000000,
        )

        self._create_server(image_uuid=uuids.image_vif_model_igb)
        self.assertIn('<model type="igb"/>', self.xml)

    def _test_boot_with_vif_model_igb_old_hypervisor(
        self, libvirt_version, qemu_version
    ):
        self.start_compute(
            hostname='compute1',
            libvirt_version=libvirt_version,
            qemu_version=qemu_version,
        )

        server = self._create_server(
            image_uuid=uuids.image_vif_model_igb, expected_state='ERROR')
        self.assertEqual(
            "No valid host was found. ", server['fault']['message'])

    def test_boot_with_vif_model_igb_old_qemu(self):
        self._test_boot_with_vif_model_igb_old_hypervisor(
            libvirt_version=9003000, qemu_version=7000000)

    def test_boot_with_vif_model_igb_old_libvirt(self):
        self._test_boot_with_vif_model_igb_old_hypervisor(
            libvirt_version=9002000, qemu_version=8000000)



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
# Copyright (c) 2010 Citrix Systems, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Handling of VM disk images.
"""

import os

from oslo_concurrency import processutils
from oslo_log import log as logging
from oslo_utils import fileutils
from oslo_utils import imageutils
from oslo_utils.imageutils import format_inspector

from nova.compute import utils as compute_utils
import nova.conf
from nova import exception
from nova.i18n import _
from nova.image import glance
import nova.privsep.qemu

LOG = logging.getLogger(__name__)

CONF = nova.conf.CONF
IMAGE_API = glance.API()


def qemu_img_info(path, format=None):
    """Return an object containing the parsed output from qemu-img info."""
    if not os.path.exists(path) and not path.startswith('rbd:'):
        raise exception.DiskNotFound(location=path)

    info = nova.privsep.qemu.unprivileged_qemu_img_info(path, format=format)
    return imageutils.QemuImgInfo(info, format='json')


def privileged_qemu_img_info(path, format=None, output_format='json'):
    """Return an object containing the parsed output from qemu-img info."""
    if not os.path.exists(path) and not path.startswith('rbd:'):
        raise exception.DiskNotFound(location=path)

    info = nova.privsep.qemu.privileged_qemu_img_info(path, format=format)
    return imageutils.QemuImgInfo(info, format='json')


def convert_image(source, dest, in_format, out_format, run_as_root=False,
                  compress=False, src_encryption=None, dest_encryption=None):
    """Convert image to other format."""
    if in_format is None:
        raise RuntimeError("convert_image without input format is a security"
                           " risk")
    _convert_image(source, dest, in_format, out_format, run_as_root,
                   compress=compress, src_encryption=src_encryption,
                   dest_encryption=dest_encryption)


def convert_image_unsafe(source, dest, out_format, run_as_root=False):
    """Convert image to other format, doing unsafe automatic input format
    detection. Do not call this function.
    """

    # NOTE: there is only 1 caller of this function:
    # imagebackend.Lvm.create_image. It is not easy to fix that without a
    # larger refactor, so for the moment it has been manually audited and
    # allowed to continue. Remove this function when Lvm.create_image has
    # been fixed.
    _convert_image(source, dest, None, out_format, run_as_root)


def _convert_image(source, dest, in_format, out_format, run_as_root,
                   compress=False, src_encryption=None, dest_encryption=None):
    try:
        with compute_utils.disk_ops_semaphore:
            if not run_as_root:
                nova.privsep.qemu.unprivileged_convert_image(
                    source, dest, in_format, out_format, CONF.instances_path,
                    compress, src_encryption=src_encryption,
                    dest_encryption=dest_encryption)
            else:
                nova.privsep.qemu.convert_image(
                    source, dest, in_format, out_format, CONF.instances_path,
                    compress, src_encryption=src_encryption,
                    dest_encryption=dest_encryption)

    except processutils.ProcessExecutionError as exp:
        msg = (_("Unable to convert image to %(format)s: %(exp)s") %
               {'format': out_format, 'exp': exp})
        raise exception.ImageUnacceptable(image_id=source, reason=msg)


def fetch(context, image_href, path, trusted_certs=None):
    with fileutils.remove_path_on_error(path):
        with compute_utils.disk_ops_semaphore:
            IMAGE_API.download(context, image_href, dest_path=path,
                               trusted_certs=trusted_certs)


def get_info(context, image_href):
    return IMAGE_API.get(context, image_href)


def check_vmdk_image(image_id, data):
    # Check some rules about VMDK files. Specifically we want to make
    # sure that the "create-type" of the image is one that we allow.
    # Some types of VMDK files can reference files outside the disk
    # image and we do not want to allow those for obvious reasons.

    types = CONF.compute.vmdk_allowed_types

    if not len(types):
        LOG.warning('Refusing to allow VMDK image as vmdk_allowed_'
                    'types is empty')
        msg = _('Invalid VMDK create-type specified')
        raise exception.ImageUnacceptable(image_id=image_id, reason=msg)

    try:
        create_type = data.format_specific['data']['create-type']
    except KeyError:
        msg = _('Unable to determine VMDK create-type')
        raise exception.ImageUnacceptable(image_id=image_id, reason=msg)

    if create_type not in CONF.compute.vmdk_allowed_types:
        LOG.warning('Refusing to process VMDK file with create-type of %r '
                    'which is not in allowed set of: %s', create_type,
                    ','.join(CONF.compute.vmdk_allowed_types))
        msg = _('Invalid VMDK create-type specified')
        raise exception.ImageUnacceptable(image_id=image_id, reason=msg)


def get_image_format(path):
    with open(path, 'rb') as f:
        wrapper = format_inspector.InspectWrapper(f)
        try:
            while f.peek():
                wrapper.read(4096)
                if wrapper.formats:
                    break
        finally:
            wrapper.close()

    try:
        return wrapper.format
    except format_inspector.ImageFormatError:
        format_names = set(str(x) for x in wrapper.formats)
        if format_names == {'iso', 'gpt'}:
            # If iso+gpt, we choose the iso because bootable-as-block ISOs
            # can legitimately have a GPT bootloader in front.
            LOG.debug('Detected %s as ISO+GPT, allowing as ISO', path)
            return [x for x in wrapper.formats if str(x) == 'iso'][0]
        # Any other case of multiple formats is an error
        raise


def do_image_deep_inspection(img, image_href, path):
    ami_formats = ('ami', 'aki', 'ari')
    disk_format = img['disk_format']
    try:
        # NOTE(danms): Use our own cautious inspector module to make sure
        # the image file passes safety checks.
        # See https://bugs.launchpad.net/nova/+bug/2059809 for details.

        # Make sure we have a format inspector for the claimed format, else
        # it is something we do not support and must reject. AMI is excluded.
        if (disk_format not in ami_formats and
                not format_inspector.get_inspector(disk_format)):
            raise exception.ImageUnacceptable(
                image_id=image_href,
                reason=_('Image not in a supported format'))

        inspector = get_image_format(path)
        inspector.safety_check()

        # Images detected as gpt but registered as raw are legacy "whole disk"
        # formats, which we continue to allow for now.
        # AMI formats can be other things, so don't obsess over this
        # requirement for them. Otherwise, make sure our detection agrees
        # with glance.
        if disk_format == 'raw' and str(inspector) == 'gpt':
            LOG.debug('Image %s registered as raw, but detected as gpt',
                      image_href)
        elif disk_format not in ami_formats and str(inspector) != disk_format:
            # If we detected the image as something other than glance claimed,
            # we abort.
            LOG.warning('Image %s expected to be %s but detected as %s',
                        image_href, disk_format, str(inspector))
            raise exception.ImageUnacceptable(
                image_id=image_href,
                reason=_('Image content does not match disk_format'))
    except format_inspector.SafetyCheckFailed as e:
        LOG.error('Image %s failed safety check: %s', image_href, e)
        raise exception.ImageUnacceptable(
            image_id=image_href,
            reason=(_('Image does not pass safety check')))
    except format_inspector.ImageFormatError:
        # If the inspector we chose based on the image's metadata does not
        # think the image is the proper format, we refuse to use it.
        raise exception.ImageUnacceptable(
            image_id=image_href,
            reason=_('Image content does not match disk_format'))
    except Exception:
        raise exception.ImageUnacceptable(
            image_id=image_href,
            reason=_('Image not in a supported format'))
    if disk_format in ('iso',) + ami_formats:
        # ISO or AMI image passed safety check; qemu will treat this as raw
        # from here so return the expected formats it will find.
        disk_format = 'raw'
    return disk_format


def fetch_to_raw(context, image_href, path, trusted_certs=None):
    path_tmp = "%s.part" % path
    fetch(context, image_href, path_tmp, trusted_certs)

    with fileutils.remove_path_on_error(path_tmp):
        if not CONF.workarounds.disable_deep_image_inspection:
            # If we're doing deep inspection, we take the determined format
            # from it.
            img = IMAGE_API.get(context, image_href)
            force_format = do_image_deep_inspection(img, image_href, path_tmp)
        else:
            force_format = None

        # Only run qemu-img after we have done deep inspection (if enabled).
        # If it was not enabled, we will let it detect the format.
        data = qemu_img_info(path_tmp)
        fmt = data.file_format
        if fmt is None:
            raise exception.ImageUnacceptable(
                reason=_("'qemu-img info' parsing failed."),
                image_id=image_href)
        elif force_format is not None and fmt != force_format:
            # Format inspector and qemu-img must agree on the format, else
            # we reject. This will catch VMDK some variants that we don't
            # explicitly support because qemu will identify them as such
            # and we will not.
            LOG.warning('Image %s detected by qemu as %s but we expected %s',
                        image_href, fmt, force_format)
            raise exception.ImageUnacceptable(
                image_id=image_href,
                reason=_('Image content does not match disk_format'))

        backing_file = data.backing_file
        if backing_file is not None:
            raise exception.ImageUnacceptable(image_id=image_href,
                reason=(_("fmt=%(fmt)s backed by: %(backing_file)s") %
                        {'fmt': fmt, 'backing_file': backing_file}))

        try:
            data_file = data.format_specific['data']['data-file']
        except (KeyError, TypeError, AttributeError):
            data_file = None
        if data_file is not None:
            raise exception.ImageUnacceptable(image_id=image_href,
                reason=(_("fmt=%(fmt)s has data-file: %(data_file)s") %
                        {'fmt': fmt, 'data_file': data_file}))

        if fmt == 'vmdk':
            check_vmdk_image(image_href, data)

        if fmt != "raw" and CONF.force_raw_images:
            staged = "%s.converted" % path
            LOG.debug("%s was %s, converting to raw", image_href, fmt)
            with fileutils.remove_path_on_error(staged):
                try:
                    convert_image(path_tmp, staged, fmt, 'raw')
                except exception.ImageUnacceptable as exp:
                    # re-raise to include image_href
                    raise exception.ImageUnacceptable(image_id=image_href,
                        reason=_("Unable to convert image to raw: %(exp)s")
                        % {'exp': exp})

                os.unlink(path_tmp)

                data = qemu_img_info(staged)
                if data.file_format != "raw":
                    raise exception.ImageUnacceptable(image_id=image_href,
                        reason=_("Converted to raw, but format is now %s") %
                        data.file_format)

                os.rename(staged, path)
        else:
            os.rename(path_tmp, path)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Handles all requests relating to shares + manila.
"""

from dataclasses import dataclass
import functools
from typing import Optional

from openstack import exceptions as sdk_exc
from oslo_log import log as logging

import nova.conf
from nova import exception
from nova import utils

CONF = nova.conf.CONF
LOG = logging.getLogger(__name__)
MIN_SHARE_FILE_SYSTEM_MICROVERSION = "2.82"


def _manilaclient(context, admin=False):
    """Constructs a manila client object for making API requests.

    :return: An openstack.proxy.Proxy object for the specified service_type.
    :raise: ConfGroupForServiceTypeNotFound If no conf group name could be
            found for the specified service_type.
    :raise: ServiceUnavailable if the service is down
    """

    return utils.get_sdk_adapter(
        "shared-file-system",
        admin=admin,
        check_service=True,
        context=context,
        shared_file_system_api_version=MIN_SHARE_FILE_SYSTEM_MICROVERSION,
        global_request_id=context.global_id
    )


@dataclass(frozen=True)
class Share():
    id: str
    size: int
    availability_zone: Optional[str]
    created_at: str
    status: str
    name: Optional[str]
    description: Optional[str]
    project_id: str
    snapshot_id: Optional[str]
    share_network_id: Optional[str]
    share_proto: str
    export_location: str
    metadata: dict
    share_type: Optional[str]
    is_public: bool

    @classmethod
    def from_manila_share(cls, manila_share, export_location):
        return cls(
            id=manila_share.id,
            size=manila_share.size,
            availability_zone=manila_share.availability_zone,
            created_at=manila_share.created_at,
            status=manila_share.status,
            name=manila_share.name,
            description=manila_share.description,
            project_id=manila_share.project_id,
            snapshot_id=manila_share.snapshot_id,
            share_network_id=manila_share.share_network_id,
            share_proto=manila_share.share_protocol,
            export_location=export_location,
            metadata=manila_share.metadata,
            share_type=manila_share.share_type,
            is_public=manila_share.is_public,
        )


@dataclass(frozen=True)
class Access():
    id: str
    access_level: str
    state: str
    access_type: str
    access_to: str
    access_key: Optional[str]

    @classmethod
    def from_manila_access(cls, manila_access):
        return cls(
            id=manila_access.id,
            access_level=manila_access.access_level,
            state=manila_access.state,
            access_type=manila_access.access_type,
            access_to=manila_access.access_to,
            access_key= getattr(manila_access, 'access_key', None)
        )

    @classmethod
    def from_dict(cls, manila_access):
        return cls(
            id=manila_access['id'],
            access_level=manila_access['access_level'],
            state=manila_access['state'],
            access_type=manila_access['access_type'],
            access_to=manila_access['access_to'],
            access_key=manila_access['access_key'],
        )


def translate_sdk_exception(method):
    """Transforms a manila exception but keeps its traceback intact."""
    @functools.wraps(method)
    def wrapper(self, *args, **kwargs):
        try:
            res = method(self, *args, **kwargs)
        except (exception.ServiceUnavailable,
                exception.ConfGroupForServiceTypeNotFound) as exc:
            raise exception.ManilaConnectionFailed(reason=str(exc)) from exc
        except (sdk_exc.BadRequestException) as exc:
            raise exception.InvalidInput(reason=str(exc)) from exc
        except (sdk_exc.ForbiddenException) as exc:
            raise exception.Forbidden(str(exc)) from exc
        return res
    return wrapper


def translate_share_exception(method):
    """Transforms the exception for the share but keeps its traceback intact.
    """

    def wrapper(self, *args, **kwargs):
        try:
            res = method(self, *args, **kwargs)
        except (sdk_exc.ResourceNotFound) as exc:
            raise exception.ShareNotFound(
                share_id=args[1], reason=exc) from exc
        except (sdk_exc.BadRequestException) as exc:
            raise exception.ShareNotFound(
                share_id=args[1], reason=exc) from exc
        return res
    return translate_sdk_exception(wrapper)


def translate_allow_exception(method):
    """Transforms the exception for allow but keeps its traceback intact.
    """

    def wrapper(self, *args, **kwargs):
        try:
            res = method(self, *args, **kwargs)
        except (sdk_exc.BadRequestException) as exc:
            raise exception.ShareAccessGrantError(
                share_id=args[1], reason=exc) from exc
        except (sdk_exc.ResourceNotFound) as exc:
            raise exception.ShareNotFound(
                share_id=args[1], reason=exc) from exc
        return res
    return translate_sdk_exception(wrapper)


def translate_deny_exception(method):
    """Transforms the exception for deny but keeps its traceback intact.
    """

    def wrapper(self, *args, **kwargs):
        try:
            res = method(self, *args, **kwargs)
        except (sdk_exc.BadRequestException) as exc:
            raise exception.ShareAccessRemovalError(
                share_id=args[1], reason=exc) from exc
        except (sdk_exc.ResourceNotFound) as exc:
            raise exception.ShareNotFound(
                share_id=args[1], reason=exc) from exc
        return res
    return translate_sdk_exception(wrapper)


class API(object):
    """API for interacting with the share manager."""

    @translate_share_exception
    def get(self, context, share_id):
        """Get the details about a share given its ID.

        :param share_id: the id of the share to get
        :raises: ShareNotFound if the share_id specified is not available.
        :returns: Share object.
        """

        def filter_export_locations(export_locations):
            # Return the preferred path otherwise choose the first one
            paths = []
            try:
                for export_location in export_locations:
                    if export_location.is_preferred:
                        return export_location.path
                    else:
                        paths.append(export_location.path)
                return paths[0]
            except (IndexError, NameError):
                return None

        client = _manilaclient(context, admin=False)
        LOG.debug("Get share id:'%s' data from manila", share_id)
        share = client.get_share(share_id)
        export_locations = client.export_locations(share.id)
        export_location = filter_export_locations(export_locations)

        return Share.from_manila_share(share, export_location)

    @translate_share_exception
    def get_access(
        self,
        context,
        share_id,
        access_type,
        access_to,
    ):
        """Get share access

        :param share_id: the id of the share to get
        :param access_type: the type of access ("ip", "cert", "user")
        :param access_to: ip:cidr or cert:cn or user:group or user name
        :raises: ShareNotFound if the share_id specified is not available.
        :returns: Access object or None if there is no access granted to this
            share.
        """

        LOG.debug("Get share access id for share id:'%s'",
                  share_id)
        access_list = _manilaclient(
            context, admin=True).access_rules(share_id)

        for access in access_list:
            if (
                access.access_type == access_type and
                access.access_to == access_to
            ):
                return Access.from_manila_access(access)
        return None

    @translate_allow_exception
    def allow(
        self,
        context,
        share_id,
        access_type,
        access_to,
        access_level,
    ):
        """Allow share access

        :param share_id: the id of the share
        :param access_type: the type of access ("ip", "cert", "user")
        :param access_to: ip:cidr or cert:cn or user:group or user name
        :param access_level: "ro" for read only or "rw" for read/write
        :raises: ShareNotFound if the share_id specified is not available.
        :raises: BadRequest if the share already exists.
        :raises: ShareAccessGrantError if the answer from manila allow API is
            not the one expected.
        """

        def check_manila_access_response(access):
            if not (
                isinstance(access, Access) and
                access.access_type == access_type and
                access.access_to == access_to and
                access.access_level == access_level
            ):
                raise exception.ShareAccessGrantError(share_id=share_id)

        LOG.debug("Allow host access to share id:'%s'",
                  share_id)

        access = _manilaclient(context, admin=True).create_access_rule(
            share_id,
            access_type=access_type,
            access_to=access_to,
            access_level=access_level,
            lock_visibility=True,
            lock_deletion=True,
            lock_reason="Lock by nova",
        )

        access = Access.from_manila_access(access)
        check_manila_access_response(access)
        return access

    @translate_deny_exception
    def deny(
        self,
        context,
        share_id,
        access_type,
        access_to,
    ):
        """Deny share access
        :param share_id: the id of the share
        :param access_type: the type of access ("ip", "cert", "user")
        :param access_to: ip:cidr or cert:cn or user:group or user name
        :raises: ShareAccessNotFound if the access_id specified is not
            available.
        :raises: ShareAccessRemovalError if the manila deny API does not
            respond with a status code 202.
        """

        access = self.get_access(
            context,
            share_id,
            access_type,
            access_to,
            )

        if access:
            client = _manilaclient(context, admin=True)
            LOG.debug("Deny host access to share id:'%s'", share_id)
            resp = client.delete_access_rule(
                access.id, share_id, unrestrict=True
            )
            if resp.status_code != 202:
                raise exception.ShareAccessRemovalError(
                    share_id=share_id, reason=resp.reason
                )
        else:
            raise exception.ShareAccessNotFound(share_id=share_id)

    def has_access(self, context, share_id, access_type, access_to):
        """Helper method to check if a policy is applied to a share
        :param context: The request context.
        :param share_id: the id of the share
        :param access_type: the type of access ("ip", "cert", "user")
        :param access_to: ip:cidr or cert:cn or user:group or user name
        :returns: boolean, true means the policy is applied.
        """
        access = self.get_access(
            context,
            share_id,
            access_type,
            access_to
        )
        return access is not None and access.state == 'active'



# Copyright 2014 NEC Corporation.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# NOTE(stephenfin): This schema is intentionally empty since the action has
# been removed
reset_network = {}

# TODO(stephenfin): Restrict the value to 'null' in a future API version
inject_network_info = {
    'type': 'object',
    'properties': {
        'injectNetworkInfo': {},
    },
    'required': ['injectNetworkInfo'],
    'additionalProperties': False,
}

reset_state = {
    'type': 'object',
    'properties': {
        'os-resetState': {
            'type': 'object',
            'properties': {
                'state': {
                    'type': 'string',
                    'enum': ['active', 'error'],
                },
            },
            'required': ['state'],
            'additionalProperties': False,
        },
    },
    'required': ['os-resetState'],
    'additionalProperties': False,
}

inject_network_info_response = {
    'type': 'null',
}

reset_state_response = {
    'type': 'null',
}



# Copyright 2016 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

spice_opt_group = cfg.OptGroup('spice',
        title="SPICE console features",
        help="""
SPICE console feature allows you to connect to a guest virtual machine.
SPICE is a replacement for fairly limited VNC protocol.

Following requirements must be met in order to use SPICE:

* Virtualization driver must be libvirt
* spice.enabled set to True
* vnc.enabled set to False
* update html5proxy_base_url
* update server_proxyclient_address
""")

CLI_OPTS = [
    cfg.HostAddressOpt('html5proxy_host',
        default='0.0.0.0',
        help="""
IP address or a hostname on which the ``nova-spicehtml5proxy`` service
listens for incoming requests.

Related options:

* This option depends on the ``html5proxy_base_url`` option.
  The ``nova-spicehtml5proxy`` service must be listening on a host that is
  accessible from the HTML5 client.
"""),
    cfg.PortOpt('html5proxy_port',
        default=6082,
        help="""
Port on which the ``nova-spicehtml5proxy`` service listens for incoming
requests.

Related options:

* This option depends on the ``html5proxy_base_url`` option.
  The ``nova-spicehtml5proxy`` service must be listening on a port that is
  accessible from the HTML5 client.
""")
]

ALL_OPTS = [
    cfg.BoolOpt('enabled',
        default=False,
        help="""
Enable SPICE related features.

Related options:

* VNC must be explicitly disabled to get access to the SPICE console. Set the
  enabled option to False in the [vnc] section to disable the VNC console.
"""),
    cfg.BoolOpt('agent_enabled',
        default=True,
        help="""
Enable the SPICE guest agent support on the instances.

The Spice agent works with the Spice protocol to offer a better guest console
experience. However, the Spice console can still be used without the Spice
Agent. With the Spice agent installed the following features are enabled:

* Copy & Paste of text and images between the guest and client machine
* Automatic adjustment of resolution when the client screen changes - e.g.
  if you make the Spice console full screen the guest resolution will adjust to
  match it rather than letterboxing.
* Better mouse integration - The mouse can be captured and released without
  needing to click inside the console or press keys to release it. The
  performance of mouse movement is also improved.
"""),
    cfg.StrOpt('image_compression',
        advanced=True,
        choices=[
            ('auto_glz', 'enable image compression mode to choose between glz '
             'and quic algorithm, based on image properties'),
            ('auto_lz', 'enable image compression mode to choose between lz '
             'and quic algorithm, based on image properties'),
            ('quic', 'enable image compression based on the SFALIC algorithm'),
            ('glz', 'enable image compression using lz with history based '
             'global dictionary'),
            ('lz', 'enable image compression with the Lempel-Ziv algorithm'),
            ('off', 'disable image compression')
        ],
        help="""
Configure the SPICE image compression (lossless).
"""),
    cfg.StrOpt('jpeg_compression',
        advanced=True,
        choices=[
            ('auto', 'enable JPEG image compression automatically'),
            ('never', 'disable JPEG image compression'),
            ('always', 'enable JPEG image compression')
        ],
        help="""
Configure the SPICE wan image compression (lossy for slow links).
"""),
    cfg.StrOpt('zlib_compression',
        advanced=True,
        choices=[
            ('auto', 'enable zlib image compression automatically'),
            ('never', 'disable zlib image compression'),
            ('always', 'enable zlib image compression')
        ],
        help="""
Configure the SPICE wan image compression (lossless for slow links).
"""),
    cfg.BoolOpt('playback_compression',
        advanced=True,
        help="""
Enable the SPICE audio stream compression (using celt).
"""),
    cfg.StrOpt('streaming_mode',
        advanced=True,
        choices=[
            ('filter', 'SPICE server adds additional filters to decide if '
             'video streaming should be activated'),
            ('all', 'any fast-refreshing window can be encoded into a video '
             'stream'),
            ('off', 'no video detection and (lossy) compression is performed')
        ],
        help="""
Configure the SPICE video stream detection and (lossy) compression.
"""),
    cfg.URIOpt('html5proxy_base_url',
        default='http://127.0.0.1:6082/spice_auto.html',
        help="""
Location of the SPICE HTML5 console proxy.

End user would use this URL to connect to the `nova-spicehtml5proxy``
service. This service will forward request to the console of an instance.

In order to use SPICE console, the service ``nova-spicehtml5proxy`` should be
running. This service is typically launched on the controller node.

Possible values:

* Must be a valid URL of the form: ``http://host:port/spice_auto.html``
  where host is the node running ``nova-spicehtml5proxy`` and the port is
  typically 6082. Consider not using default value as it is not well defined
  for any real deployment.

Related options:

* This option depends on ``html5proxy_host`` and ``html5proxy_port`` options.
  The access URL returned by the compute node must have the host
  and port where the ``nova-spicehtml5proxy`` service is listening.
"""),
    cfg.URIOpt('spice_direct_proxy_base_url',
        default='http://127.0.0.1:13002/nova',
        help="""
Location of a SPICE protocol native console proxy.

A user can retrieve a virt-viewer style .vv connection configuration file by
accessing this URL with the attached token when a console is created.

Possible values:

* Must be a valid URL of the form: ``http://host:port/nova`` where host is the
  node running the SPICE protocol native proxy and the port is typically 13002.
  Note that the port component is optional if you are using the default port
  for HTTP or HTTPS. Consider not using the default value as it is not well
  defined for any real deployment.
"""),
    cfg.StrOpt('server_listen',
        default='127.0.0.1',
        help="""
The  address where the SPICE server running on the instances should listen.

Typically, the ``nova-spicehtml5proxy`` proxy client runs on the controller
node and connects over the private network to this address on the compute
node(s).

Possible values:

* IP address to listen on.
"""),
    cfg.StrOpt('server_proxyclient_address',
        default='127.0.0.1',
        help="""
The address used by ``nova-spicehtml5proxy`` client to connect to instance
console.

Typically, the ``nova-spicehtml5proxy`` proxy client runs on the
controller node and connects over the private network to this address on the
compute node(s).

Possible values:

* Any valid IP address on the compute node.

Related options:

* This option depends on the ``server_listen`` option.
  The proxy client must be able to access the address specified in
  ``server_listen`` using the value of this option.
"""),
    cfg.BoolOpt('require_secure',
        default=False,
        help="""
Whether to require secure TLS connections to SPICE consoles.

If you're providing direct access to SPICE consoles instead of using the HTML5
proxy, you may wish those connections to be encrypted. If so, set this value to
True.

Note that use of secure consoles requires that you setup TLS certificates on
each hypervisor.

Possible values:

* False: console traffic is not encrypted.
* True: console traffic is required to be protected by TLS.
"""),
]

ALL_OPTS.extend(CLI_OPTS)


def register_opts(conf):
    conf.register_opts(ALL_OPTS, group=spice_opt_group)


def register_cli_opts(conf):
    conf.register_cli_opts(CLI_OPTS, group=spice_opt_group)


def list_opts():
    return {spice_opt_group: ALL_OPTS}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# The driver is not designed to function as a volume for libvirt. Rather,
# its purpose is to facilitate the mounting of cephfs shares exposed by manila
# on compute, enabling the shares to be accessed and shared using virtiofs.
# This is the reason why get_config and extend_volume are not implemented.

import nova.conf
from nova.virt.libvirt.volume import fs

CONF = nova.conf.CONF


class LibvirtCEPHFSVolumeDriver(fs.LibvirtMountedFileSystemVolumeDriver):
    """Class implements libvirt part of volume driver for CEPHFS."""

    def __init__(self, connection):
        super(LibvirtCEPHFSVolumeDriver, self).__init__(connection, 'ceph')

    def _get_mount_point_base(self):
        return CONF.libvirt.ceph_mount_point_base

    def get_config(self, connection_info, disk_info):
        raise NotImplementedError()

    def _mount_options(self, connection_info):
        options = []
        conn_options = connection_info['data'].get('options')
        if CONF.libvirt.ceph_mount_options is not None:
            options.extend(CONF.libvirt.ceph_mount_options)
            if conn_options:
                options.extend(conn_options)
            options = [','.join(options)]
            options.insert(0, '-o')
        else:
            if conn_options:
                options.extend(['-o', ','.join(conn_options)])

        return options

    def extend_volume(self, connection_info, instance, requested_size):
        raise NotImplementedError()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from nova.api.validation import parameter_types

# NOTE(stephenfin): These schemas are incomplete but won't be enhanced further
# since these APIs have been removed

show_query = {
    'type': 'object',
    'properties': {},
    'additionalProperties': True,
}

index_query = {
    'type': 'object',
    'properties': {
        # NOTE(stephenfin): We never validated these and we're not going to add
        # that validation now.
        # field filters
        'name': {},
        'status': {},
        'changes-since': {},
        'server': {},
        'type': {},
        'minRam': {},
        'minDisk': {},
        # pagination filters
        'limit': parameter_types.multi_params(
            parameter_types.positive_integer),
        'page_size': parameter_types.multi_params(
            parameter_types.positive_integer),
        'marker': {},
        'offset': parameter_types.multi_params(
            parameter_types.positive_integer),
    },
    'patternProperties': {
        '^property-.*$': {},
    },
    'additionalProperties': True,
}

detail_query = index_query



#   Copyright 2011 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from nova.api.openstack.compute import admin_actions
from nova.tests.unit.api.openstack.compute import admin_only_action_common


class AdminActionsTest(admin_only_action_common.CommonTests):

    def setUp(self):
        super().setUp()
        self.controller = admin_actions.AdminActionsController()
        self.compute_api = self.controller.compute_api

    def test_actions(self):
        actions = ['_inject_network_info']
        method_translations = {'_inject_network_info': 'inject_network_info'}

        self._test_actions(
            actions, method_translations,
            {'_inject_network_info': {'injectNetworkInfo': None}})

    def test_actions_with_non_existed_instance(self):
        actions = ['_inject_network_info']
        self._test_actions_with_non_existed_instance(
            actions,
            {'_inject_network_info': {'injectNetworkInfo': None}})

    def test_actions_with_locked_instance(self):
        actions = ['_inject_network_info']
        method_translations = {'_inject_network_info': 'inject_network_info'}

        self._test_actions_with_locked_instance(
            actions, method_translations,
            {'_inject_network_info': {'injectNetworkInfo': None}})



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from nova.api.openstack.compute import quota_sets
from nova import exception
from nova.policies import quota_sets as policies
from nova.tests.unit.api.openstack import fakes
from nova.tests.unit.policies import base


class QuotaSetsPolicyTest(base.BasePolicyTest):
    """Test Quota Sets APIs policies with all possible context.
    This class defines the set of context with different roles
    which are allowed and not allowed to pass the policy checks.
    With those set of context, it will call the API operation and
    verify the expected behaviour.
    """

    def setUp(self):
        super(QuotaSetsPolicyTest, self).setUp()
        self.controller = quota_sets.QuotaSetsController()
        self.controller._validate_quota_limit = mock.MagicMock()
        self.req = fakes.HTTPRequest.blank('')
        self.project_id = self.req.environ['nova.context'].project_id

        # With legacy rule all admin is able to update or revert their quota
        # to default or get other project quota.
        self.project_admin_authorized_contexts = set([
            self.legacy_admin_context, self.system_admin_context,
            self.project_admin_context])
        # With legacy rule, everyone is able to get their own quota.
        self.project_reader_authorized_contexts = set([
            self.legacy_admin_context, self.system_admin_context,
            self.project_admin_context,
            self.system_member_context, self.system_reader_context,
            self.system_foo_context, self.project_member_context,
            self.project_reader_context, self.project_foo_context,
            self.other_project_member_context,
            self.other_project_reader_context])
        # Everyone is able to get the default quota
        self.everyone_authorized_contexts = set([
            self.legacy_admin_context, self.system_admin_context,
            self.project_admin_context,
            self.system_member_context, self.system_reader_context,
            self.system_foo_context, self.project_member_context,
            self.project_reader_context, self.project_foo_context,
            self.other_project_member_context,
            self.other_project_reader_context])

    @mock.patch('nova.quota.QUOTAS.get_project_quotas')
    @mock.patch('nova.quota.QUOTAS.get_settable_quotas')
    def test_update_quota_sets_policy(self, mock_update, mock_get):
        rule_name = policies.POLICY_ROOT % 'update'
        body = {'quota_set': {
                    'instances': 50,
                    'cores': 50}
               }
        for cxtx in self.project_admin_authorized_contexts:
            req = fakes.HTTPRequest.blank('')
            req.environ['nova.context'] = cxtx
            self.controller.update(req, cxtx.project_id, body=body)
        for cxtx in (self.all_contexts -
                set(self.project_admin_authorized_contexts)):
            req = fakes.HTTPRequest.blank('')
            req.environ['nova.context'] = cxtx
            exc = self.assertRaises(
                exception.PolicyNotAuthorized, self.controller.update,
                req, cxtx.project_id, body=body)
            self.assertEqual(
                "Policy doesn't allow %s to be performed." % rule_name,
                exc.format_message())

    @mock.patch('nova.objects.Quotas.destroy_all_by_project')
    def test_delete_quota_sets_policy(self, mock_delete):
        rule_name = policies.POLICY_ROOT % 'delete'
        for cxtx in self.project_admin_authorized_contexts:
            req = fakes.HTTPRequest.blank('')
            req.environ['nova.context'] = cxtx
            self.controller.delete(req, cxtx.project_id)
        for cxtx in (self.all_contexts -
                set(self.project_admin_authorized_contexts)):
            req = fakes.HTTPRequest.blank('')
            req.environ['nova.context'] = cxtx
            exc = self.assertRaises(
                exception.PolicyNotAuthorized, self.controller.delete,
                req, cxtx.project_id)
            self.assertEqual(
                "Policy doesn't allow %s to be performed." % rule_name,
                exc.format_message())

    @mock.patch('nova.quota.QUOTAS.get_defaults')
    def test_default_quota_sets_policy(self, mock_default):
        rule_name = policies.POLICY_ROOT % 'defaults'
        self.common_policy_auth(self.everyone_authorized_contexts,
                                rule_name,
                                self.controller.defaults,
                                self.req, self.project_id)

    @mock.patch('nova.quota.QUOTAS.get_project_quotas')
    def test_detail_quota_sets_policy(self, mock_get):
        rule_name = policies.POLICY_ROOT % 'detail'
        self.common_policy_auth(self.project_admin_authorized_contexts,
                                rule_name,
                                self.controller.detail,
                                self.req, 'try-other-project')
        # Check if project reader or higher roles are able to get
        # their own quota
        for cxtx in self.project_reader_authorized_contexts:
            req = fakes.HTTPRequest.blank('')
            req.environ['nova.context'] = cxtx
            self.controller.detail(req, cxtx.project_id)

    @mock.patch('nova.quota.QUOTAS.get_project_quotas')
    def test_show_quota_sets_policy(self, mock_get):
        rule_name = policies.POLICY_ROOT % 'show'
        self.common_policy_auth(self.project_admin_authorized_contexts,
                                rule_name,
                                self.controller.show,
                                self.req, 'try-other-project')
        # Check if project reader or higher roles are able to get
        # their own quota
        for cxtx in self.project_reader_authorized_contexts:
            req = fakes.HTTPRequest.blank('')
            req.environ['nova.context'] = cxtx
            self.controller.show(req, cxtx.project_id)


class QuotaSetsNoLegacyNoScopePolicyTest(QuotaSetsPolicyTest):
    """Test QuotaSets APIs policies with no legacy deprecated rules
    and no scope checks.

    """

    without_deprecated_rules = True

    def setUp(self):
        super(QuotaSetsNoLegacyNoScopePolicyTest, self).setUp()
        # Even with no legacy rule, because any admin requesting
        # update/revert quota for their own project will be allowed.
        # And any admin will be able to get other project quota.
        self.project_admin_authorized_contexts = [
            self.legacy_admin_context, self.system_admin_context,
            self.project_admin_context]
        # With no legacy rule, other project and foo role will not be
        # able to get the quota.
        self.project_reader_authorized_contexts = [
            self.legacy_admin_context, self.system_admin_context,
            self.project_admin_context,
            self.system_member_context, self.system_reader_context,
            self.project_member_context,
            self.project_reader_context]


class QuotaSetsScopeTypePolicyTest(QuotaSetsPolicyTest):
    """Test Quota Sets APIs policies with system scope enabled.
    This class set the nova.conf [oslo_policy] enforce_scope to True
    so that we can switch on the scope checking on oslo policy side.
    It defines the set of context with scoped token
    which are allowed and not allowed to pass the policy checks.
    With those set of context, it will run the API operation and
    verify the expected behaviour.
    """

    def setUp(self):
        super(QuotaSetsScopeTypePolicyTest, self).setUp()
        self.flags(enforce_scope=True, group="oslo_policy")

        # With scope enabled, system users will be disallowed.
        self.reduce_set('project_admin_authorized', set([
            self.legacy_admin_context,
            self.project_admin_context]))
        self.reduce_set('project_reader_authorized',
                        self.all_project_contexts)
        self.everyone_authorized_contexts = self.all_project_contexts


class QuotaSetsScopeTypeNoLegacyPolicyTest(QuotaSetsScopeTypePolicyTest):
    """Test Quota Sets APIs policies with system scope enabled,
    and no more deprecated rules that allow the legacy admin API to
    access system APIs.
    """
    without_deprecated_rules = True

    def setUp(self):
        super(QuotaSetsScopeTypeNoLegacyPolicyTest, self).setUp()
        # With scope enabled and no legacy, system and
        # non-reader/member users are disallowed.
        self.reduce_set('project_reader_authorized',
                        self.all_project_contexts -
                        set([self.project_foo_context]))



# Copyright 2016 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_policy import policy

from nova.policies import base


POLICY_ROOT = 'os_compute_api:os-aggregates:%s'
NEW_POLICY_ROOT = 'compute:aggregates:%s'


aggregates_policies = [
    policy.DocumentedRuleDefault(
        name=POLICY_ROOT % 'set_metadata',
        check_str=base.ADMIN,
        description="Create or replace metadata for an aggregate",
        operations=[
            {
                'path': '/os-aggregates/{aggregate_id}/action (set_metadata)',
                'method': 'POST'
            }
        ],
        scope_types=['project']),
    policy.DocumentedRuleDefault(
        name=POLICY_ROOT % 'add_host',
        check_str=base.ADMIN,
        description="Add a host to an aggregate",
        operations=[
            {
                'path': '/os-aggregates/{aggregate_id}/action (add_host)',
                'method': 'POST'
            }
        ],
        scope_types=['project']),
    policy.DocumentedRuleDefault(
        name=POLICY_ROOT % 'create',
        check_str=base.ADMIN,
        description="Create an aggregate",
        operations=[
            {
                'path': '/os-aggregates',
                'method': 'POST'
            }
        ],
        scope_types=['project']),
    policy.DocumentedRuleDefault(
        name=POLICY_ROOT % 'remove_host',
        check_str=base.ADMIN,
        description="Remove a host from an aggregate",
        operations=[
            {
                'path': '/os-aggregates/{aggregate_id}/action (remove_host)',
                'method': 'POST'
            }
        ],
        scope_types=['project']),
    policy.DocumentedRuleDefault(
        name=POLICY_ROOT % 'update',
        check_str=base.ADMIN,
        description="Update name and/or availability zone for an aggregate",
        operations=[
            {
                'path': '/os-aggregates/{aggregate_id}',
                'method': 'PUT'
            }
        ],
        scope_types=['project']),
    policy.DocumentedRuleDefault(
        name=POLICY_ROOT % 'index',
        check_str=base.ADMIN,
        description="List all aggregates",
        operations=[
            {
                'path': '/os-aggregates',
                'method': 'GET'
            }
        ],
        scope_types=['project']),
    policy.DocumentedRuleDefault(
        name=POLICY_ROOT % 'delete',
        check_str=base.ADMIN,
        description="Delete an aggregate",
        operations=[
            {
                'path': '/os-aggregates/{aggregate_id}',
                'method': 'DELETE'
            }
        ],
        scope_types=['project']),
    policy.DocumentedRuleDefault(
        name=POLICY_ROOT % 'show',
        check_str=base.ADMIN,
        description="Show details for an aggregate",
        operations=[
            {
                'path': '/os-aggregates/{aggregate_id}',
                'method': 'GET'
            }
        ],
        scope_types=['project']),
    policy.DocumentedRuleDefault(
        name=NEW_POLICY_ROOT % 'images',
        check_str=base.ADMIN,
        description="Request image caching for an aggregate",
        operations=[
            {
                'path': '/os-aggregates/{aggregate_id}/images',
                'method': 'POST'
            }
        ],
        scope_types=['project']),
]


def list_rules():
    return aggregates_policies



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import fixtures
from oslo_utils.fixture import uuidsentinel as uuids

from nova.api.openstack.compute import server_metadata
from nova.policies import server_metadata as policies
from nova.tests.unit.api.openstack import fakes
from nova.tests.unit import fake_instance
from nova.tests.unit.policies import base


class ServerMetadataPolicyTest(base.BasePolicyTest):
    """Test Server Metadata APIs policies with all possible context.
    This class defines the set of context with different roles
    which are allowed and not allowed to pass the policy checks.
    With those set of context, it will call the API operation and
    verify the expected behaviour.
    """

    def setUp(self):
        super(ServerMetadataPolicyTest, self).setUp()
        self.controller = server_metadata.ServerMetadataController()
        self.req = fakes.HTTPRequest.blank('')
        self.mock_get = self.useFixture(
            fixtures.MockPatch('nova.api.openstack.common.get_instance')).mock
        self.instance = fake_instance.fake_instance_obj(
                self.project_member_context,
                id=1, uuid=uuids.fake_id, project_id=self.project_id)
        self.mock_get.return_value = self.instance

        # With legacy rule and no scope checks, all admin, project members
        # project reader or other project role(because legacy rule allow server
        # owner- having same project id and no role check) is able to create,
        # update, and delete the server metadata.
        self.project_member_authorized_contexts = [
            self.legacy_admin_context, self.system_admin_context,
            self.project_admin_context, self.project_member_context,
            self.project_reader_context, self.project_foo_context]
        # and they can get their own server metadata.
        self.project_reader_authorized_contexts = (
            self.project_member_authorized_contexts)

    @mock.patch('nova.compute.api.API.get_instance_metadata')
    def test_index_server_Metadata_policy(self, mock_get):
        rule_name = policies.POLICY_ROOT % 'index'
        self.common_policy_auth(self.project_reader_authorized_contexts,
                                rule_name,
                                self.controller.index,
                                self.req, self.instance.uuid)

    @mock.patch('nova.compute.api.API.get_instance_metadata')
    def test_show_server_Metadata_policy(self, mock_get):
        rule_name = policies.POLICY_ROOT % 'show'
        mock_get.return_value = {'key9': 'value'}
        self.common_policy_auth(self.project_reader_authorized_contexts,
                                rule_name,
                                self.controller.show,
                                self.req, self.instance.uuid, 'key9')

    @mock.patch('nova.compute.api.API.update_instance_metadata')
    def test_create_server_Metadata_policy(self, mock_quota):
        rule_name = policies.POLICY_ROOT % 'create'
        self.common_policy_auth(self.project_member_authorized_contexts,
                                rule_name,
                                self.controller.create,
                                self.req, self.instance.uuid,
                                body={"metadata": {"key9": "value9"}})

    @mock.patch('nova.compute.api.API.update_instance_metadata')
    def test_update_server_Metadata_policy(self, mock_quota):
        rule_name = policies.POLICY_ROOT % 'update'
        self.common_policy_auth(self.project_member_authorized_contexts,
                                rule_name,
                                self.controller.update,
                                self.req, self.instance.uuid, 'key9',
                                body={"meta": {"key9": "value9"}})

    @mock.patch('nova.compute.api.API.update_instance_metadata')
    def test_update_all_server_Metadata_policy(self, mock_quota):
        rule_name = policies.POLICY_ROOT % 'update_all'
        self.common_policy_auth(self.project_member_authorized_contexts,
                                rule_name,
                                self.controller.update_all,
                                self.req, self.instance.uuid,
                                body={"metadata": {"key9": "value9"}})

    @mock.patch('nova.compute.api.API.get_instance_metadata')
    @mock.patch('nova.compute.api.API.delete_instance_metadata')
    def test_delete_server_Metadata_policy(self, mock_delete, mock_get):
        rule_name = policies.POLICY_ROOT % 'delete'
        mock_get.return_value = {'key9': 'value'}
        self.common_policy_auth(self.project_member_authorized_contexts,
                                rule_name,
                                self.controller.delete,
                                self.req, self.instance.uuid, 'key9')


class ServerMetadataNoLegacyNoScopePolicyTest(ServerMetadataPolicyTest):
    """Test Server Metadata APIs policies with no legacy deprecated rules
    and no scope checks which means new defaults only.

    """

    without_deprecated_rules = True

    def setUp(self):
        super(ServerMetadataNoLegacyNoScopePolicyTest, self).setUp()
        # With no legacy rule, legacy admin loose power.
        self.project_member_authorized_contexts = (
            self.project_member_or_admin_with_no_scope_no_legacy)
        self.project_reader_authorized_contexts = (
            self.project_reader_or_admin_with_no_scope_no_legacy)


class ServerMetadataScopeTypePolicyTest(ServerMetadataPolicyTest):
    """Test Server Metadata APIs policies with system scope enabled.
    This class set the nova.conf [oslo_policy] enforce_scope to True
    so that we can switch on the scope checking on oslo policy side.
    It defines the set of context with scoped token
    which are allowed and not allowed to pass the policy checks.
    With those set of context, it will run the API operation and
    verify the expected behaviour.
    """

    def setUp(self):
        super(ServerMetadataScopeTypePolicyTest, self).setUp()
        self.flags(enforce_scope=True, group="oslo_policy")
        # With Scope enable, system users no longer allowed.
        self.project_member_authorized_contexts = (
            self.project_m_r_or_admin_with_scope_and_legacy)
        self.project_reader_authorized_contexts = (
            self.project_m_r_or_admin_with_scope_and_legacy)


class ServerMetadataScopeTypeNoLegacyPolicyTest(
        ServerMetadataScopeTypePolicyTest):
    """Test Server Metadata APIs policies with system scope enabled,
    and no more deprecated rules that allow the legacy admin API to
    access system APIs.
    """
    without_deprecated_rules = True

    def setUp(self):
        super(ServerMetadataScopeTypeNoLegacyPolicyTest, self).setUp()
        # With no legacy and scope enable, only project admin, member,
        # and reader will be able to allowed operation on server metadata.
        self.project_member_authorized_contexts = (
            self.project_member_or_admin_with_scope_no_legacy)
        self.project_reader_authorized_contexts = (
            self.project_reader_or_admin_with_scope_no_legacy)



#   Copyright 2013 Rackspace Hosting
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

"""The shelved mode extension."""

from oslo_log import log as logging
from webob import exc

from nova.api.openstack import api_version_request
from nova.api.openstack import common
from nova.api.openstack.compute.schemas import shelve as schema
from nova.api.openstack import wsgi
from nova.api import validation
from nova.compute import api as compute
from nova import exception
from nova.policies import shelve as shelve_policies

LOG = logging.getLogger(__name__)


class ShelveController(wsgi.Controller):
    def __init__(self):
        super(ShelveController, self).__init__()
        self.compute_api = compute.API()

    @wsgi.response(202)
    @wsgi.expected_errors((404, 403, 409, 400))
    @wsgi.action('shelve')
    @validation.schema(schema.shelve)
    @validation.response_body_schema(schema.shelve_response)
    def _shelve(self, req, id, body):
        """Move an instance into shelved mode."""
        context = req.environ["nova.context"]

        instance = common.get_instance(self.compute_api, context, id)
        context.can(shelve_policies.POLICY_ROOT % 'shelve',
                    target={'user_id': instance.user_id,
                            'project_id': instance.project_id})
        try:
            self.compute_api.shelve(context, instance)
        except (
            exception.InstanceIsLocked,
            exception.UnexpectedTaskStateError,
        ) as e:
            raise exc.HTTPConflict(explanation=e.format_message())
        except exception.InstanceInvalidState as state_error:
            common.raise_http_conflict_for_instance_invalid_state(state_error,
                                                                  'shelve', id)
        except exception.ForbiddenPortsWithAccelerator as e:
            raise exc.HTTPBadRequest(explanation=e.format_message())
        except (
            exception.ForbiddenSharesNotSupported,
            exception.ForbiddenWithShare) as e:
            raise exc.HTTPConflict(explanation=e.format_message())

    @wsgi.response(202)
    @wsgi.expected_errors((400, 404, 409))
    @wsgi.action('shelveOffload')
    @validation.schema(schema.shelve_offload)
    @validation.response_body_schema(schema.shelve_offload_response)
    def _shelve_offload(self, req, id, body):
        """Force removal of a shelved instance from the compute node."""
        context = req.environ["nova.context"]
        instance = common.get_instance(self.compute_api, context, id)
        context.can(shelve_policies.POLICY_ROOT % 'shelve_offload',
                    target={'user_id': instance.user_id,
                            'project_id': instance.project_id})
        try:
            self.compute_api.shelve_offload(context, instance)
        except exception.InstanceIsLocked as e:
            raise exc.HTTPConflict(explanation=e.format_message())
        except exception.InstanceInvalidState as state_error:
            common.raise_http_conflict_for_instance_invalid_state(state_error,
                                                              'shelveOffload',
                                                              id)

        except exception.ForbiddenPortsWithAccelerator as e:
            raise exc.HTTPBadRequest(explanation=e.format_message())

    @wsgi.response(202)
    @wsgi.expected_errors((400, 403, 404, 409))
    @wsgi.action('unshelve')
    @validation.schema(schema.unshelve, '2.1', '2.76')
    # In microversion 2.77 we support specifying 'availability_zone' to
    # unshelve a server. But before 2.77 there is no request body
    # schema validation (because of body=null).
    @validation.schema(schema.unshelve_v277, '2.77', '2.90')
    # In microversion 2.91 we support specifying 'host' to
    # unshelve an instance to a specific hostself.
    # 'availability_zone' = None is supported as well to unpin the
    # availability zone of an instance bonded to this availability_zone
    @validation.schema(schema.unshelve_v291, '2.91')
    @validation.response_body_schema(schema.unshelve_response)
    def _unshelve(self, req, id, body):
        """Restore an instance from shelved mode."""
        context = req.environ["nova.context"]
        instance = common.get_instance(self.compute_api, context, id)
        context.can(
            shelve_policies.POLICY_ROOT % 'unshelve',
            target={'project_id': instance.project_id}
        )

        unshelve_args = {}

        unshelve_dict = body.get('unshelve')
        support_az = api_version_request.is_supported(
            req, '2.77')
        support_host = api_version_request.is_supported(
            req, '2.91')
        if unshelve_dict:
            if support_az and 'availability_zone' in unshelve_dict:
                unshelve_args['new_az'] = (
                    unshelve_dict['availability_zone']
                )
            if support_host:
                unshelve_args['host'] = unshelve_dict.get('host')

        try:
            self.compute_api.unshelve(
                context,
                instance,
                **unshelve_args,
            )
        except (
            exception.InstanceIsLocked,
            exception.UnshelveInstanceInvalidState,
            exception.UnshelveHostNotInAZ,
            exception.MismatchVolumeAZException,
        ) as e:
            raise exc.HTTPConflict(explanation=e.format_message())
        except exception.InstanceInvalidState as state_error:
            common.raise_http_conflict_for_instance_invalid_state(
                state_error, 'unshelve', id)
        except (
            exception.InvalidRequest,
            exception.ExtendedResourceRequestOldCompute,
            exception.ComputeHostNotFound,
        ) as e:
            raise exc.HTTPBadRequest(explanation=e.format_message())
        except exception.OverQuota as e:
            raise exc.HTTPForbidden(explanation=e.format_message())



# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""The deferred instance delete extension."""

import webob

from nova.api.openstack import common
from nova.api.openstack.compute.schemas import deferred_delete as schema
from nova.api.openstack import wsgi
from nova.api import validation
from nova.compute import api as compute
from nova import exception
from nova.policies import deferred_delete as dd_policies


class DeferredDeleteController(wsgi.Controller):
    def __init__(self):
        super(DeferredDeleteController, self).__init__()
        self.compute_api = compute.API()

    @wsgi.response(202)
    @wsgi.expected_errors((403, 404, 409))
    @wsgi.action('restore')
    @validation.schema(schema.restore)
    @validation.response_body_schema(schema.restore_response)
    def _restore(self, req, id, body):
        """Restore a previously deleted instance."""
        context = req.environ["nova.context"]
        instance = common.get_instance(self.compute_api, context, id)
        context.can(dd_policies.BASE_POLICY_NAME % 'restore',
                    target={'project_id': instance.project_id})
        try:
            self.compute_api.restore(context, instance)
        except exception.OverQuota as error:
            raise webob.exc.HTTPForbidden(explanation=error.format_message())
        except exception.InstanceInvalidState as state_error:
            common.raise_http_conflict_for_instance_invalid_state(state_error,
                    'restore', id)

    @wsgi.response(202)
    @wsgi.expected_errors((404, 409))
    @wsgi.action('forceDelete')
    @validation.schema(schema.force_delete)
    @validation.response_body_schema(schema.force_delete_response)
    def _force_delete(self, req, id, body):
        """Force delete of instance before deferred cleanup."""
        context = req.environ["nova.context"]
        instance = common.get_instance(self.compute_api, context, id)
        context.can(dd_policies.BASE_POLICY_NAME % 'force',
                    target={'user_id': instance.user_id,
                            'project_id': instance.project_id})
        try:
            self.compute_api.force_delete(context, instance)
        except exception.InstanceNotFound as e:
            raise webob.exc.HTTPNotFound(explanation=e.format_message())
        except exception.InstanceIsLocked as e:
            raise webob.exc.HTTPConflict(explanation=e.format_message())



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy

# NOTE(stephenfin): These schemas are intentionally empty since these APIs are
# deprecated proxy APIs

create = {}
add_interface = {}
remove_interface = {}

index_query = {}
show_query = {}

_node = {
    'type': 'object',
    'properties': {
        'cpus': {'type': 'string'},
        'disk_gb': {'type': 'string'},
        'host': {'const': 'IRONIC MANAGED'},
        'id': {'type': 'string', 'format': 'uuid'},
        'interfaces': {
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': {
                    'address': {'type': 'string'},
                },
            },
        },
        'memory_mb': {'type': 'string'},
        'task_state': {'type': 'string'},
    },
    'required': [
        'cpus',
        'disk_gb',
        'host',
        'id',
        'interfaces',
        'memory_mb',
        'task_state',
    ],
    'additionalProperties': False,
}
_node_detail = copy.deepcopy(_node)
_node_detail['properties']['instance_uuid'] = {
    'type': 'string', 'format': 'uuid'
}
_node_detail['required'].append('instance_uuid')

index_response = {
    'type': 'object',
    'properties': {
        'nodes': {
            'type': 'array',
            'items': copy.deepcopy(_node),
        },
    },
    'required': ['nodes'],
    'additionalProperties': False,
}

show_response = {
    'type': 'object',
    'properties': {
        'node': copy.deepcopy(_node_detail),
    },
    'required': ['node'],
    'additionalProperties': False,
}

create_response = {
    'type': 'null',
}

delete_response = {
    'type': 'null',
}

add_interface_response = {
    'type': 'null',
}

remove_interface_response = {
    'type': 'null',
}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Functions to address filesystem calls, particularly sysfs."""

import functools
import os
import time

from oslo_log import log as logging

from nova import exception

LOG = logging.getLogger(__name__)
SYS = '/sys'
RETRY_LIMIT = 5


# a retry decorator to handle EBUSY
def retry_if_busy(func):
    """Decorator to retry a function if it raises DeviceBusy.

    This decorator will retry the function RETRY_LIMIT=5 times if it raises
    DeviceBusy. It will sleep for 1 second on the first retry, 2 seconds on
    the second retry, and so on, up to RETRY_LIMIT seconds. If the function
    still raises DeviceBusy after RETRY_LIMIT retries, the exception will be
    raised.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        for i in range(RETRY_LIMIT):
            try:
                return func(*args, **kwargs)
            except exception.DeviceBusy as e:
                # if we have retried RETRY_LIMIT times, raise the exception
                # otherwise, sleep and retry, i is 0-based so we need
                # to add 1 to it
                count = i + 1
                if count < RETRY_LIMIT:
                    LOG.debug(
                        f"File {e.kwargs['file_path']} is busy, "
                        f"sleeping {count} seconds before retrying")
                    time.sleep(count)
                    continue
                raise
    return wrapper

# NOTE(bauzas): this method is deliberately not wrapped in a privsep entrypoint


@retry_if_busy
def read_sys(path: str) -> str:
    """Reads the content of a file in the sys filesystem.

    :param path: relative or absolute. If relative, will be prefixed by /sys.
    :returns: contents of that file.
    :raises: nova.exception.FileNotFound if we can't read that file.
    :raises: nova.exception.DeviceBusy if the file is busy.
    """
    try:
        # The path can be absolute with a /sys prefix but that's fine.
        with open(os.path.join(SYS, path), mode='r') as data:
            return data.read()
    except OSError as exc:
        # errno 16 is EBUSY, which means the file is busy.
        if exc.errno == 16:
            raise exception.DeviceBusy(file_path=path) from exc
        # convert permission denied to file not found
        raise exception.FileNotFound(file_path=path) from exc
    except ValueError as exc:
        raise exception.FileNotFound(file_path=path) from exc


# NOTE(bauzas): this method is deliberately not wrapped in a privsep entrypoint
# In order to correctly use it, you need to decorate the caller with a specific
# privsep entrypoint.
@retry_if_busy
def write_sys(path: str, data: str) -> None:
    """Writes the content of a file in the sys filesystem with data.

    :param path: relative or absolute. If relative, will be prefixed by /sys.
    :param data: the data to write.
    :returns: contents of that file.
    :raises: nova.exception.FileNotFound if we can't write that file.
    :raises: nova.exception.DeviceBusy if the file is busy.
    """
    try:
        # The path can be absolute with a /sys prefix but that's fine.
        with open(os.path.join(SYS, path), mode='w') as fd:
            fd.write(data)
    except OSError as exc:
        # errno 16 is EBUSY, which means the file is busy.
        if exc.errno == 16:
            raise exception.DeviceBusy(file_path=path) from exc
        # convert permission denied to file not found
        raise exception.FileNotFound(file_path=path) from exc
    except ValueError as exc:
        raise exception.FileNotFound(file_path=path) from exc



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Image Properties Weigher. Weigh hosts by the image metadata properties
related to the existing instances.

A positive value will favor hosts with the same image properties (packing
strategy) while a negative value will follow a spread strategy that will favor
hosts not already having instances with those image properties.
The default value of the multiplier is 0, which disables the weigher.
"""

import nova.conf
from nova import exception
from nova import objects
from nova.scheduler import utils
from nova.scheduler import weights

CONF = nova.conf.CONF


class ImagePropertiesWeigher(weights.BaseHostWeigher):
    def __init__(self):
        self._parse_setting()

    def _parse_setting(self):
        self.setting = dict(utils.parse_options(
            CONF.filter_scheduler.image_props_weight_setting,
            sep='=', converter=float,
            name="filter_scheduler.image_props_weight_setting"))

    def weight_multiplier(self, host_state):
        """Override the weight multiplier."""
        return utils.get_weight_multiplier(
            host_state, 'image_props_weight_multiplier',
            CONF.filter_scheduler.image_props_weight_multiplier)

    def _weigh_object(self, host_state, request_spec):
        """Higher weights win.  We want to choose hosts with the more common
           existing image properties that are used by instances by default.
           If you want to spread instances with the same properties between
           hosts, change the multiplier value to a negative number.
        """

        weight = 0.0

        # Disable this weigher if we don't use it as it's a bit costly.
        if CONF.filter_scheduler.image_props_weight_multiplier == 0.0:
            return weight

        # request_spec is a RequestSpec object which can have its image
        # field set to None
        if request_spec.image:
            # List values aren't hashable so we need to stringify them.
            requested_props = {(key, f"{value}") for key, value in
                request_spec.image.properties.to_dict().items()}
        else:
            requested_props = set()

        existing_props = []

        insts = objects.InstanceList(objects=host_state.instances.values())
        # system_metadata isn't loaded yet, let's do this.
        insts.fill_metadata()

        for inst in insts:
            try:
                props = {(key, str(value)) for key, value in
                         inst.image_meta.properties.to_dict().items()
                         } if inst.image_meta else set()
            except exception.InstanceNotFound:
                # the host state can be a bit stale as the instance could no
                # longer exist on the host if the instance deletion arrives
                # before the scheduler gets the RPC message of the deletion
                props = set()
            # We want to unpack the set of tuples as items to the total list
            # of properties we need to compare.
            existing_props.extend(tup for tup in props)

        common_props = requested_props & set(existing_props)

        for (prop, value) in common_props:
            if self.setting:
                # Calculate the weigh for each property by what was set
                # If it wasn't defined, then don't weigh this property.
                weight += self.setting.get(
                            prop, 0.0) * existing_props.count((prop, value))
            else:
                # By default, all properties are weighed evenly.
                weight += existing_props.count((prop, value))
        return weight



# Copyright (c) 2011-2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Pluggable Weighing support
"""

import abc

from oslo_log import log as logging

from nova import loadables


LOG = logging.getLogger(__name__)


def normalize(weight_list, minval=None, maxval=None):
    """Normalize the values in a list between 0 and 1.0.

    The normalization is made regarding the lower and upper values present in
    weight_list. If the minval and/or maxval parameters are set, these values
    will be used instead of the minimum and maximum from the list.

    If all the values are equal, they are normalized to 0.
    """

    if not weight_list:
        return ()

    if maxval is None:
        maxval = max(weight_list)

    if minval is None:
        minval = min(weight_list)

    maxval = float(maxval)
    minval = float(minval)

    if minval == maxval:
        return [0] * len(weight_list)

    range_ = maxval - minval
    return ((i - minval) / range_ for i in weight_list)


class WeighedObject(object):
    """Object with weight information."""

    def __init__(self, obj, weight):
        self.obj = obj
        self.weight = weight

    def __repr__(self):
        return "<WeighedObject '%s': %s>" % (self.obj, self.weight)


class BaseWeigher(metaclass=abc.ABCMeta):
    """Base class for pluggable weighers.

    The attributes maxval and minval can be specified to set up the maximum
    and minimum values for the weighed objects. These values will then be
    taken into account in the normalization step, instead of taking the values
    from the calculated weights.
    """

    minval = None
    maxval = None

    def weight_multiplier(self, host_state):
        """How weighted this weigher should be.

        Override this method in a subclass, so that the returned value is
        read from a configuration option to permit operators specify a
        multiplier for the weigher. If the host is in an aggregate, this
        method of subclass can read the ``weight_multiplier`` from aggregate
        metadata of ``host_state``, and use it to overwrite multiplier
        configuration.

        :param host_state: The HostState object.
        """
        return 1.0

    @abc.abstractmethod
    def _weigh_object(self, obj, weight_properties):
        """Weigh an specific object."""

    def weigh_objects(self, weighed_obj_list, weight_properties):
        """Weigh multiple objects.

        Override in a subclass if you need access to all objects in order
        to calculate weights. Do not modify the weight of an object here,
        just return a list of weights.
        """
        # Calculate the weights
        weights = []
        for obj in weighed_obj_list:
            weight = self._weigh_object(obj.obj, weight_properties)

            # don't let the weight go beyond the defined max/min
            if self.minval is not None:
                weight = max(weight, self.minval)
            if self.maxval is not None:
                weight = min(weight, self.maxval)

            weights.append(weight)

        return weights


class BaseWeightHandler(loadables.BaseLoader):
    object_class = WeighedObject

    def get_weighed_objects(self, weighers, obj_list, weighing_properties):
        """Return a sorted (descending), normalized list of WeighedObjects."""
        weighed_objs = [self.object_class(obj, 0.0) for obj in obj_list]

        if len(weighed_objs) <= 1:
            return weighed_objs

        for weigher in weighers:
            weights = weigher.weigh_objects(weighed_objs, weighing_properties)

            LOG.debug(
                "%s: raw weights %s",
                weigher.__class__.__name__,
                {(obj.obj.host, obj.obj.nodename): weight
                 for obj, weight in zip(weighed_objs, weights)}
            )

            # Normalize the weights
            weights = list(
                normalize(
                    weights, minval=weigher.minval, maxval=weigher.maxval))

            LOG.debug(
                "%s: normalized weights %s",
                weigher.__class__.__name__,
                {(obj.obj.host, obj.obj.nodename): weight
                 for obj, weight in zip(weighed_objs, weights)}
            )

            log_data = {}

            for i, weight in enumerate(weights):
                obj = weighed_objs[i]
                multiplier = weigher.weight_multiplier(obj.obj)
                weigher_score = multiplier * weight
                obj.weight += weigher_score

                log_data[(obj.obj.host, obj.obj.nodename)] = (
                    f"{multiplier} * {weight}")

            LOG.debug(
                "%s: score (multiplier * weight) %s",
                weigher.__class__.__name__,
                {name: log for name, log in log_data.items()}
            )

        return sorted(weighed_objs, key=lambda x: x.weight, reverse=True)



# Copyright 2014 IBM Corporation.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy


# TODO(stephenfin): Remove additionalProperties in a future API version
index_query = {
    'type': 'object',
    'properties': {},
    'additionalProperties': True,
}

detail_query = index_query

index_response = {
    'type': 'object',
    'properties': {
        'availabilityZoneInfo': {
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': {
                    'hosts': {'type': 'null'},
                    'zoneName': {'type': 'string'},
                    'zoneState': {
                        'type': 'object',
                        'properties': {
                            'available': {'type': 'boolean'},
                        },
                    },
                },
                'required': ['hosts', 'zoneName', 'zoneState'],
                'additionalProperties': False,
            },
        },
    },
    'required': ['availabilityZoneInfo'],
    'additionalProperties': False,
}

detail_response = copy.deepcopy(index_response)
detail_response['properties']['availabilityZoneInfo']['items']['properties']['hosts'] = {  # noqa: E501
    'type': ['null', 'object'],
    'patternProperties': {
        '^.+$': {
            'type': 'object',
            'patternProperties': {
                '^.+$': {
                    'type': 'object',
                    'properties': {
                        'active': {'type': 'boolean'},
                        'available': {'type': 'boolean'},
                        'updated_at': {
                            'oneOf': [
                                {'type': 'null'},
                                {'type': 'string', 'format': 'date-time'},
                            ],
                        },
                    },
                    'required': ['active', 'available', 'updated_at'],
                    'additionalProperties': False,
                },
            },
            'additionalProperties': False,
        },
    },
    'additionalProperties': False,
}



# Copyright 2013 NEC Corporation.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from nova.api.validation import parameter_types


add_tenant_access = {
    'type': 'object',
    'properties': {
        'addTenantAccess': {
            'type': 'object',
            'properties': {
                'tenant': {
                    # defined from project_id in instance_type_projects table
                    'type': 'string', 'minLength': 1, 'maxLength': 255,
                },
            },
            'required': ['tenant'],
            'additionalProperties': False,
        },
    },
    'required': ['addTenantAccess'],
    'additionalProperties': False,
}

remove_tenant_access = {
    'type': 'object',
    'properties': {
        'removeTenantAccess': {
            'type': 'object',
            'properties': {
                'tenant': {
                    # defined from project_id in instance_type_projects table
                    'type': 'string', 'minLength': 1, 'maxLength': 255,
                },
            },
            'required': ['tenant'],
            'additionalProperties': False,
        },
    },
    'required': ['removeTenantAccess'],
    'additionalProperties': False,
}

# TODO(stephenfin): Remove additionalProperties in a future API version
index_query = {
    'type': 'object',
    'properties': {},
    'additionalProperties': True,
}

_common_response = {
    'type': 'object',
    'properties': {
        'flavor_access': {
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': {
                    'flavor_id': {'type': 'string'},
                    'tenant_id': parameter_types.project_id,
                },
                'required': ['flavor_id', 'tenant_id'],
                'additionalProperties': True,
            },
        },
    },
    'required': ['flavor_access'],
    'additionalProperties': True,
}

index_response = copy.deepcopy(_common_response)

add_tenant_access_response = copy.deepcopy(_common_response)

remove_tenant_access_response = copy.deepcopy(_common_response)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from keystoneauth1 import loading as ks_loading
from oslo_config import cfg

from nova.conf import utils as confutils

DEFAULT_SERVICE_TYPE = 'shared-file-system'

manila_group = cfg.OptGroup(
    'manila',
    title='Manila Options',
    help="Configuration options for the share-file-system service")

manila_opts = [
    cfg.IntOpt('share_apply_policy_timeout',
               default=10,
               help="""
Timeout period for share policy application.

Maximum duration to await a response from the Manila service for the
application of a share policy before experiencing a timeout.
0 means do not wait (0s).

Possible values:

* A positive integer or 0 (default value is 10).
"""),
]


def register_opts(conf):
    conf.register_group(manila_group)
    conf.register_opts(manila_opts, group=manila_group)
    ks_loading.register_session_conf_options(conf, manila_group.name)
    ks_loading.register_auth_conf_options(conf, manila_group.name)

    confutils.register_ksa_opts(conf, manila_group, DEFAULT_SERVICE_TYPE)


def list_opts():
    return {
        manila_group.name: (
            manila_opts +
            ks_loading.get_session_conf_options() +
            ks_loading.get_auth_common_conf_options() +
            ks_loading.get_auth_plugin_conf_options('v3password'))
    }



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import webob

from oslo_db import exception as db_exc
from oslo_utils import uuidutils

from nova.api.openstack import common
from nova.api.openstack.compute.schemas import server_shares as schema
from nova.api.openstack.compute.views import server_shares
from nova.api.openstack import wsgi
from nova.api import validation
from nova.compute import api as compute
from nova.compute import vm_states
from nova import context as nova_context
from nova import exception
from nova import objects
from nova.objects import fields
from nova.policies import server_shares as ss_policies
from nova.share import manila
from nova.virt import hardware as hw


def _get_instance_mapping(context, server_id):
    try:
        return objects.InstanceMapping.get_by_instance_uuid(context, server_id)
    except exception.InstanceMappingNotFound as e:
        raise webob.exc.HTTPNotFound(explanation=e.format_message())


class ServerSharesController(wsgi.Controller):
    _view_builder_class = server_shares.ViewBuilder

    def __init__(self):
        super(ServerSharesController, self).__init__()
        self.compute_api = compute.API()
        self.manila = manila.API()

    def _get_instance_from_server_uuid(self, context, server_id):
        instance = common.get_instance(self.compute_api, context, server_id)
        return instance

    def _check_instance_in_valid_state(self, context, server_id, action):
        instance = self._get_instance_from_server_uuid(context, server_id)
        if (
            (action == "create share" and
             instance.vm_state not in vm_states.STOPPED) or
            (action == "delete share" and
             instance.vm_state not in vm_states.STOPPED and
             instance.vm_state not in vm_states.ERROR)
        ):
            exc = exception.InstanceInvalidState(
                attr="vm_state",
                instance_uuid=instance.uuid,
                state=instance.vm_state,
                method=action,
            )
            common.raise_http_conflict_for_instance_invalid_state(
                exc, action, server_id
            )
        return instance

    @wsgi.Controller.api_version("2.97")
    @wsgi.response(200)
    @wsgi.expected_errors((400, 403, 404))
    @validation.query_schema(schema.index_query)
    @validation.response_body_schema(schema.index_response)
    def index(self, req, server_id):
        context = req.environ["nova.context"]
        # Get instance mapping to query the required cell database
        im = _get_instance_mapping(context, server_id)
        context.can(ss_policies.POLICY_ROOT % 'index',
                    target={'project_id': im.project_id})

        with nova_context.target_cell(context, im.cell_mapping) as cctxt:
            # Ensure the instance exists
            self._get_instance_from_server_uuid(cctxt, server_id)
            db_shares = objects.ShareMappingList.get_by_instance_uuid(
                cctxt, server_id
            )

        return self._view_builder._list_view(db_shares)

    @wsgi.Controller.api_version("2.97")
    @wsgi.response(201)
    @wsgi.expected_errors((400, 403, 404, 409))
    @validation.schema(schema.create, min_version='2.97')
    @validation.response_body_schema(schema.show_response)
    def create(self, req, server_id, body):
        def _try_create_share_mapping(context, share_mapping):
            """Block the request if the share is already created.
            Prevent race conditions of requests that would hit the
            share_mapping.create() almost at the same time.
            Prevent user from using the same tag twice on the same instance.
            """
            try:
                objects.ShareMapping.get_by_instance_uuid_and_share_id(context,
                    share_mapping.instance_uuid, share_mapping.share_id
                )
                raise exception.ShareMappingAlreadyExists(
                    share_id=share_mapping.share_id, tag=share_mapping.tag
                )
            except exception.ShareNotFound:
                pass

            try:
                share_mapping.create()
            except db_exc.DBDuplicateEntry:
                raise exception.ShareMappingAlreadyExists(
                    share_id=share_mapping.share_id, tag=share_mapping.tag
                )

        def _check_manila_share(manila_share_data):
            """Check that the targeted share in manila has
            correct export location, status 'available' and a supported
            protocol.
            """
            if manila_share_data.status != 'available':
                raise exception.ShareStatusIncorect(
                    share_id=share_id, status=manila_share_data.status
                )

            if manila_share_data.export_location is None:
                raise exception.ShareMissingExportLocation(share_id=share_id)

            if (
                manila_share_data.share_proto
                not in fields.ShareMappingProto.ALL
            ):
                raise exception.ShareProtocolNotSupported(
                    share_proto=manila_share_data.share_proto
                )

        context = req.environ["nova.context"]
        # Get instance mapping to query the required cell database
        im = _get_instance_mapping(context, server_id)
        context.can(
            ss_policies.POLICY_ROOT % 'create',
            target={'project_id': im.project_id}
        )

        share_dict = body['share']
        share_id = share_dict.get('share_id')
        with nova_context.target_cell(context, im.cell_mapping) as cctxt:
            instance = self._check_instance_in_valid_state(
                cctxt,
                server_id,
                "create share"
            )

            try:
                hw.check_shares_supported(cctxt, instance)

                manila_share_data = self.manila.get(cctxt, share_id)
                _check_manila_share(manila_share_data)

                share_mapping = objects.ShareMapping(cctxt)
                share_mapping.uuid = uuidutils.generate_uuid()
                share_mapping.instance_uuid = server_id
                share_mapping.share_id = manila_share_data.id
                share_mapping.status = fields.ShareMappingStatus.ATTACHING
                share_mapping.tag = share_dict.get('tag', manila_share_data.id)
                share_mapping.export_location = (
                    manila_share_data.export_location)
                share_mapping.share_proto = manila_share_data.share_proto

                _try_create_share_mapping(cctxt, share_mapping)
                self.compute_api.allow_share(cctxt, instance, share_mapping)

                view = self._view_builder._show_view(cctxt, share_mapping)

            except (exception.ShareNotFound) as e:
                raise webob.exc.HTTPNotFound(explanation=e.format_message())
            except (exception.ShareStatusIncorect) as e:
                raise webob.exc.HTTPConflict(explanation=e.format_message())
            except (exception.ShareMissingExportLocation) as e:
                raise webob.exc.HTTPConflict(explanation=e.format_message())
            except (exception.ShareProtocolNotSupported) as e:
                raise webob.exc.HTTPConflict(explanation=e.format_message())
            except (exception.ShareMappingAlreadyExists) as e:
                raise webob.exc.HTTPConflict(explanation=e.format_message())
            except (exception.ForbiddenSharesNotSupported) as e:
                raise webob.exc.HTTPForbidden(explanation=e.format_message())
            except (exception.ForbiddenSharesNotConfiguredCorrectly) as e:
                raise webob.exc.HTTPConflict(explanation=e.format_message())

        return view

    @wsgi.Controller.api_version("2.97")
    @wsgi.response(200)
    @wsgi.expected_errors((400, 403, 404))
    @validation.query_schema(schema.show_query)
    @validation.response_body_schema(schema.show_response)
    def show(self, req, server_id, id):
        context = req.environ["nova.context"]
        # Get instance mapping to query the required cell database
        im = _get_instance_mapping(context, server_id)
        context.can(
            ss_policies.POLICY_ROOT % 'show',
            target={'project_id': im.project_id}
        )

        with nova_context.target_cell(context, im.cell_mapping) as cctxt:
            try:
                # Ensure the instance exists
                self._get_instance_from_server_uuid(cctxt, server_id)
                share = objects.ShareMapping.get_by_instance_uuid_and_share_id(
                    cctxt,
                    server_id,
                    id
                )

                view = self._view_builder._show_view(cctxt, share)

            except (exception.ShareNotFound) as e:
                raise webob.exc.HTTPNotFound(explanation=e.format_message())

        return view

    @wsgi.Controller.api_version("2.97")
    @wsgi.response(200)
    @wsgi.expected_errors((400, 403, 404, 409))
    def delete(self, req, server_id, id):
        context = req.environ["nova.context"]
        # Get instance mapping to query the required cell database
        im = _get_instance_mapping(context, server_id)
        context.can(
            ss_policies.POLICY_ROOT % 'delete',
            target={'project_id': im.project_id}
        )

        with nova_context.target_cell(context, im.cell_mapping) as cctxt:
            instance = self._check_instance_in_valid_state(
                cctxt,
                server_id,
                "delete share"
            )
            try:
                # Ensure the instance exists
                self._get_instance_from_server_uuid(cctxt, server_id)
                share_mapping = (
                    objects.ShareMapping.get_by_instance_uuid_and_share_id(
                        cctxt, server_id, id
                    )
                )

                share_mapping.status = fields.ShareMappingStatus.DETACHING
                share_mapping.save()
                self.compute_api.deny_share(cctxt, instance, share_mapping)

            except (exception.ShareNotFound) as e:
                raise webob.exc.HTTPNotFound(explanation=e.format_message())



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import copy
import ddt

from oslo_utils.fixture import uuidsentinel as uuids

from nova.tests.functional.libvirt import base


@ddt.ddt
class TestSchedulingForImageCPUProperty(base.ServersTestBase):
    """Regression test for bug #2062425

    reference arch list
    https://docs.openstack.org/glance/latest/admin/useful-image-properties.html#image-property-keys-and-values
    https://github.com/openstack/nova/blob/df5454021543e8aa3c73289224158c9090f0da6a/nova/scheduler/request_filter.py#L208-L216
    """
    microversion = 'latest'

    def setUp(self):
        super().setUp()

        self.flags(image_metadata_prefilter=True, group='scheduler')
        self.start_compute()

    def test_server_create_with_valid_arch(self):
        arch = "x86_64"
        cpu_arch_image = copy.deepcopy(self.glance.image1)
        cpu_arch_image['properties']['hw_architecture'] = arch
        cpu_arch_image['id'] = uuids.cpu_arch_image
        self.glance.create(None, cpu_arch_image)

        self._create_server(
            image_uuid=uuids.cpu_arch_image,
            networks='none',
            expected_state='ACTIVE'
        )

    def test_server_create_with_valid_emul_arch(self):
        emul_arch = "x86_64"
        emul_image = copy.deepcopy(self.glance.image1)
        emul_image['properties']['hw_emulation_architecture'] = emul_arch
        emul_image['id'] = uuids.emul_image
        self.glance.create(None, emul_image)

        self._create_server(
            image_uuid=uuids.emul_image,
            networks='none',
            expected_state='ACTIVE'
        )



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import fixtures
from oslo_utils.fixture import uuidsentinel as uuids

from nova.api.openstack.compute import server_password
from nova.policies import base as base_policy
from nova.policies import server_password as policies
from nova.tests.unit.api.openstack import fakes
from nova.tests.unit import fake_instance
from nova.tests.unit.policies import base


class ServerPasswordPolicyTest(base.BasePolicyTest):
    """Test Server Password APIs policies with all possible context.
    This class defines the set of context with different roles
    which are allowed and not allowed to pass the policy checks.
    With those set of context, it will call the API operation and
    verify the expected behaviour.
    """

    def setUp(self):
        super(ServerPasswordPolicyTest, self).setUp()
        self.controller = server_password.ServerPasswordController()
        self.req = fakes.HTTPRequest.blank('')
        self.mock_get = self.useFixture(
            fixtures.MockPatch('nova.api.openstack.common.get_instance')).mock
        self.instance = fake_instance.fake_instance_obj(
                self.project_member_context,
                id=1, uuid=uuids.fake_id, project_id=self.project_id,
                system_metadata={}, expected_attrs=['system_metadata'])
        self.mock_get.return_value = self.instance
        # With legacy rule and no scope checks, all admin, project members
        # project reader or other project role(because legacy rule allow server
        # owner- having same project id and no role check) is able to delete,
        # the server Password.
        self.project_member_authorized_contexts = [
            self.legacy_admin_context, self.system_admin_context,
            self.project_admin_context, self.project_member_context,
            self.project_reader_context, self.project_foo_context]
        # and they can get their own server password.
        self.project_reader_authorized_contexts = (
            self.project_member_authorized_contexts)

    @mock.patch('nova.api.metadata.password.extract_password')
    def test_index_server_password_policy(self, mock_pass):
        rule_name = policies.BASE_POLICY_NAME % 'show'
        self.common_policy_auth(self.project_reader_authorized_contexts,
                                rule_name,
                                self.controller.index,
                                self.req, self.instance.uuid)

    @mock.patch('nova.api.metadata.password.convert_password')
    def test_clear_server_password_policy(self, mock_pass):
        rule_name = policies.BASE_POLICY_NAME % 'clear'
        self.common_policy_auth(self.project_member_authorized_contexts,
                                rule_name,
                                self.controller.clear,
                                self.req, self.instance.uuid)


class ServerPasswordNoLegacyNoScopePolicyTest(ServerPasswordPolicyTest):
    """Test Server Password APIs policies with no legacy deprecated rules
    and no scope checks.

    """

    without_deprecated_rules = True
    rules_without_deprecation = {
        policies.BASE_POLICY_NAME % 'show':
            base_policy.PROJECT_READER_OR_ADMIN,
        policies.BASE_POLICY_NAME % 'clear':
            base_policy.PROJECT_MEMBER_OR_ADMIN}

    def setUp(self):
        super(ServerPasswordNoLegacyNoScopePolicyTest, self).setUp()
        # With no legacy rule, legacy admin loose power.
        self.project_member_authorized_contexts = (
            self.project_member_or_admin_with_no_scope_no_legacy)
        self.project_reader_authorized_contexts = (
            self.project_reader_or_admin_with_no_scope_no_legacy)


class ServerPasswordScopeTypePolicyTest(ServerPasswordPolicyTest):
    """Test Server Password APIs policies with system scope enabled.
    This class set the nova.conf [oslo_policy] enforce_scope to True
    so that we can switch on the scope checking on oslo policy side.
    It defines the set of context with scoped token
    which are allowed and not allowed to pass the policy checks.
    With those set of context, it will run the API operation and
    verify the expected behaviour.
    """

    def setUp(self):
        super(ServerPasswordScopeTypePolicyTest, self).setUp()
        self.flags(enforce_scope=True, group="oslo_policy")
        # With Scope enable, system users no longer allowed.
        self.project_member_authorized_contexts = (
            self.project_m_r_or_admin_with_scope_and_legacy)
        self.project_reader_authorized_contexts = (
            self.project_m_r_or_admin_with_scope_and_legacy)


class ServerPasswordScopeTypeNoLegacyPolicyTest(
        ServerPasswordScopeTypePolicyTest):
    """Test Server Password APIs policies with system scope enabled,
    and no more deprecated rules.
    """
    without_deprecated_rules = True
    rules_without_deprecation = {
        policies.BASE_POLICY_NAME % 'show':
            base_policy.PROJECT_READER_OR_ADMIN,
        policies.BASE_POLICY_NAME % 'clear':
            base_policy.PROJECT_MEMBER_OR_ADMIN}

    def setUp(self):
        super(ServerPasswordScopeTypeNoLegacyPolicyTest, self).setUp()
        # With no legacy and scope enable, only project admin, member,
        # and reader will be able to allowed operation on server password.
        self.project_member_authorized_contexts = (
            self.project_member_or_admin_with_scope_no_legacy)
        self.project_reader_authorized_contexts = (
            self.project_reader_or_admin_with_scope_no_legacy)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from nova.api.validation import parameter_types

create = {
    'title': 'Server shares',
    'type': 'object',
    'properties': {
        'share': {
            'type': 'object',
            'properties': {
                'share_id': parameter_types.share_id,
                'tag': parameter_types.share_tag,
            },
            'required': ['share_id'],
            'additionalProperties': False
        }
    },
    'required': ['share'],
    'additionalProperties': False
}

index_query = {
    'type': 'object',
    'properties': {},
    'additionalProperties': False
}

show_query = {
    'type': 'object',
    'properties': {},
    'additionalProperties': False
}

# "share": {
#   "uuid": "68ba1762-fd6d-4221-8311-f3193dd93404",
#   "export_location": "10.0.0.50:/mnt/foo",
#   "share_id": "e8debdc0-447a-4376-a10a-4cd9122d7986",
#   "status": "inactive",
#   "tag": "e8debdc0-447a-4376-a10a-4cd9122d7986"
# }

_share_response = {
    'type': 'object',
    'properties': {
        'export_location': parameter_types.share_export_location,
        'share_id': parameter_types.share_id,
        'status': parameter_types.share_status,
        'tag': parameter_types.share_tag,
        'uuid': parameter_types.share_id,
    },
    'required': ['share_id', 'status', 'tag'],
    'additionalProperties': False
}

show_response = {
    'title': 'Server share',
    'type': 'object',
    'properties': {
        'share': _share_response
    },
    'required': ['share'],
    'additionalProperties': False
}

index_response = {
    'title': 'Server shares',
    'type': 'object',
    'properties': {
        'shares': {
            'type': 'array',
            'items': _share_response
        },
    },
    'required': ['shares'],
    'additionalProperties': False
}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import fixtures
import threading

from oslo_log import log as logging

from nova.tests.functional.libvirt import base


LOG = logging.getLogger(__name__)


class TestComputeStartupInProgressEvacuation(
    base.ServersTestBase
):
    """Regression test for bug 2085975. """
    CAST_AS_CALL = False
    ADMIN_API = True
    microversion = 'latest'

    def test_compute_startup_evacuation_in_progress(self):
        """Test that when the source compute started up while an instance
        evacuation from that compute is still in progress on the destination
        then the source compute properly sends the shared storage check RPC
        call to the dest compute based on the in progress migration and not
        based on the instance.host.

        To be able to test it we need a bit of setup:
        * we need to ensure check_instance_shared_storage_local passes in
          the functional env instead of returning None and triggering skip of
          the remote check
        * we need to hook the check_instance_shared_storage RPC call to be
          able to assert its success / failure
        * we need to stop the evacuation on the dest compute at the right time
          when the migration record is already in progress and the
          instance.host is not yet set to the dest. We hook the libvirt
          driver's rebuild method to wait.

        """
        # we are expecting to hit RPC timeout so set it to low value to speed
        # the test up
        self.flags(rpc_response_timeout=1)

        self.start_compute('compute1')
        self.start_compute('compute2')

        # this call does not work in the functional env as it always returns
        # None due to the instance directory does not exist locally.
        # This means the shared storage check is skipped in functional env.
        # But we need that to happen in this test case so we mock the function
        # to return success on the local check so that the RPC call to the
        # remote end is triggered.
        # Also, we need to mock globally not just in compute1 as the compute1
        # restart will instantiate a new driver.
        self.useFixture(fixtures.MockPatch(
            'nova.virt.libvirt.LibvirtDriver.'
            'check_instance_shared_storage_local',
            return_value={"filename": "foo"}))

        # To be able to assert that shared storage check failed due to RPC
        # timeout we need to hook check_instance_shared_storage RPC call
        compute1_rpc = self.computes['compute1'].compute_rpcapi
        origi_check_shared_storage = compute1_rpc.check_instance_shared_storage
        rpc_exception = None
        rpc_target_host = None

        def wrapped_check_shared_storage(ctxt, data, instance, host=None):
            try:
                nonlocal rpc_target_host
                rpc_target_host = host
                return origi_check_shared_storage(ctxt, data, instance, host)
            except Exception as e:
                nonlocal rpc_exception
                rpc_exception = e

        self.useFixture(fixtures.MockPatch(
            'nova.compute.rpcapi.ComputeAPI.check_instance_shared_storage',
            side_effect=wrapped_check_shared_storage))

        # need to make sure the evac is stopped at target compute. This
        # simulates that the evac is slow due to e.g. a lot of volume connects
        # needs to be re-attached.
        sleeping = threading.Event()
        contd = threading.Event()

        def slow_evac(*args, **kwargs):
            sleeping.set()
            contd.wait()
            # this will trigger the normal rebuild path for the libvirt driver
            # that does not implement any specific rebuild call.
            raise NotImplementedError()

        self.useFixture(fixtures.MockPatchObject(
            self.computes['compute2'].driver, 'rebuild',
            side_effect=slow_evac))

        # Create a test server on the first compute
        server = self._create_server(host='compute1', networks='none')

        # stop the compute hosting our instance and force it down to allow
        # evacuation
        self.computes['compute1'].stop()
        service_id = self.admin_api.get_services(
            host='compute1', binary='nova-compute')[0]['id']
        self.admin_api.put_service_force_down(service_id, True)

        # request the evacuation but do not wait until it succeeds
        self.api.post_server_action(server['id'], {'evacuate': {}})
        # wait until the evac hits compute2
        sleeping.wait()

        # start up the source compute while the evacuation is still running
        self.restart_compute_service('compute1')

        # allow the rebuild to continue to ensure the test case cleanly
        # finishes
        contd.set()

        self.assertIsNone(rpc_exception)
        self.assertEqual('compute2', rpc_target_host)



# Copyright 2014 NEC Corporation.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from nova.api.validation import parameter_types

_metadata_key = '^[a-zA-Z0-9-_:. ]{1,255}$'
# NOTE(oomichi): The metadata of flavor_extraspecs should accept numbers
# as its values.
_metadata = copy.deepcopy(parameter_types.metadata)
_metadata['patternProperties'][_metadata_key]['type'] = [
    'string', 'number'
]

create = {
    'type': 'object',
    'properties': {
        'extra_specs': _metadata
    },
    'required': ['extra_specs'],
    'additionalProperties': False,
}

update = copy.deepcopy(_metadata)
update.update({
     'minProperties': 1,
     'maxProperties': 1
})

# TODO(stephenfin): Remove additionalProperties in a future API version
index_query = {
    'type': 'object',
    'properties': {},
    'additionalProperties': True,
}

# TODO(stephenfin): Remove additionalProperties in a future API version
show_query = {
    'type': 'object',
    'properties': {},
    'additionalProperties': True,
}

index_response = {
    'type': 'object',
    'properties': {
        # NOTE(stephenfin): While we accept numbers as values, we always return
        # strings
        'extra_specs': parameter_types.metadata,
    },
    'required': ['extra_specs'],
    'additionalProperties': False,
}

# NOTE(stephenfin): We return the request back to the user unmodified, meaning
# if the user provided a number key then they'll get a number key back, even
# though later requests will return a string
create_response = copy.deepcopy(create)

# NOTE(stephenfin): As above
update_response = copy.deepcopy(update)

# NOTE(stephenfin): Since we are retrieving here, we always return string keys
# (like index)
show_response = copy.deepcopy(parameter_types.metadata)

delete_response = {
    'type': 'null',
}



# Copyright 2022 StackHPC
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_limit import exception as limit_exceptions
from oslo_limit import limit
from oslo_log import log as logging

import nova.conf

LOG = logging.getLogger(__name__)
CONF = nova.conf.CONF

UNIFIED_LIMITS_DRIVER = "nova.quota.UnifiedLimitsDriver"
ENDPOINT = None


def use_unified_limits():
    return CONF.quota.driver == UNIFIED_LIMITS_DRIVER


def _endpoint():
    global ENDPOINT
    if ENDPOINT is None:
        # This is copied from oslo_limit/limit.py
        endpoint_id = CONF.oslo_limit.endpoint_id
        if not endpoint_id:
            raise ValueError("endpoint_id is not configured")
        enforcer = limit.Enforcer(lambda: None)
        ENDPOINT = enforcer.connection.get_endpoint(endpoint_id)
    return ENDPOINT


def should_enforce(exc: limit_exceptions.ProjectOverLimit) -> bool:
    """Whether the exceeded resource limit should be enforced.

    Given a ProjectOverLimit exception from oslo.limit, check whether the
    involved limit(s) should be enforced. This is needed if we need more logic
    than is available by default in oslo.limit.

    :param exc: An oslo.limit ProjectOverLimit exception instance, which
        contains a list of OverLimitInfo. Each OverLimitInfo includes a
        resource_name, limit, current_usage, and delta.
    """
    # If any exceeded limit is greater than zero, it means an explicitly set
    # limit has been enforced. And if any explicitly set limit has gone over
    # quota, the enforcement should be upheld and there is no need to consider
    # the potential for unset limits.
    if any(info.limit > 0 for info in exc.over_limit_info_list):
        return True

    # Next, if all of the exceeded limits are -1, we don't need to enforce and
    # we can avoid calling Keystone for the list of registered limits.
    #
    # A value of -1 is documented in Keystone as meaning unlimited:
    #
    # "Note
    #  The default limit of registered limit and the resource limit of project
    #  limit now are limited from -1 to 2147483647 (integer). -1 means no limit
    #  and 2147483647 is the max value for user to define limits."
    #
    # https://docs.openstack.org/keystone/latest/admin/unified-limits.html#what-is-a-limit
    #
    # but oslo.limit enforce does not treat -1 as unlimited at this time and
    # instead uses its literal integer value. We will consider any negative
    # limit value as unlimited.
    if all(info.limit < 0 for info in exc.over_limit_info_list):
        return False

    # Only resources with exceeded limits of "0" are candidates for
    # enforcement.
    #
    # A limit of "0" in the over_limit_info_list means that oslo.limit is
    # telling us the limit is 0. But oslo.limit returns 0 for two cases:
    # a) it found a limit of 0 in Keystone or b) it did not find a limit in
    # Keystone at all.
    #
    # We will need to query the list of registered limits from Keystone in
    # order to determine whether each "0" limit is case a) or case b).
    enforce_candidates = {
        info.resource_name for info in exc.over_limit_info_list
            if info.limit == 0}

    # Get a list of all the registered limits. There is not a way to filter by
    # resource names however this will do one API call whereas the alternative
    # is calling GET /registered_limits/{registered_limit_id} for each resource
    # name.
    enforcer = limit.Enforcer(lambda: None)
    registered_limits = list(enforcer.connection.registered_limits(
        service_id=_endpoint().service_id, region_id=_endpoint().region_id))

    # Make a set of resource names of the registered limits.
    have_limits_set = {limit.resource_name for limit in registered_limits}

    # If any candidates have limits set, enforce. It means at least one limit
    # has been explicitly set to 0.
    if enforce_candidates & have_limits_set:
        return True

    # The resource list will be either a require list or an ignore list.
    require_or_ignore = CONF.quota.unified_limits_resource_list

    strategy = CONF.quota.unified_limits_resource_strategy
    enforced = enforce_candidates
    if strategy == 'require':
        # Resources that are in both the candidate list and in the require list
        # should be enforced.
        enforced = enforce_candidates & set(require_or_ignore)
    elif strategy == 'ignore':
        # Resources that are in the candidate list but are not in the ignore
        # list should be enforced.
        enforced = enforce_candidates - set(require_or_ignore)
    else:
        LOG.error(
            f'Invalid strategy value: {strategy} is specified in the '
            '[quota]unified_limits_resource_strategy config option, so '
            f'enforcing for resources {enforced}')
    # Log in case we need to debug unexpected enforcement or non-enforcement.
    msg = (
        f'enforcing for resources {enforced}' if enforced else 'not enforcing')
    LOG.debug(
        f'Resources {enforce_candidates} have no registered limits set in '
        f'Keystone. [quota]unified_limits_resource_strategy is {strategy} and '
        f'[quota]unified_limits_resource_list is {require_or_ignore}, '
        f'so {msg}')
    return bool(enforced)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from unittest import mock

from oslo_serialization import jsonutils
from oslo_utils.fixture import uuidsentinel as uuids
from oslo_versionedobjects import base as ovo_base
import testtools

from nova import exception
from nova import objects
from nova.objects import fields
from nova.tests.unit.objects import test_objects


fake_instance_uuid = uuids.fake

fake_obj_numa_topology = objects.InstanceNUMATopology(
    instance_uuid=fake_instance_uuid,
    cells=[
        objects.InstanceNUMACell(
            id=0, cpuset=set(), pcpuset=set([1, 2]),
            cpuset_reserved=set([5, 6]), memory=512,
            pagesize=2048),
        objects.InstanceNUMACell(
            id=1, cpuset=set(), pcpuset=set([3, 4]),
            cpuset_reserved=set([7, 8]), memory=512,
            pagesize=2048),
    ])

fake_db_topology = {
    'created_at': None,
    'updated_at': None,
    'deleted_at': None,
    'deleted': 0,
    'id': 1,
    'instance_uuid': fake_instance_uuid,
    'numa_topology': fake_obj_numa_topology._to_json()
    }


def get_fake_obj_numa_topology(context):
    fake_obj_numa_topology_cpy = fake_obj_numa_topology.obj_clone()
    fake_obj_numa_topology_cpy._context = context
    return fake_obj_numa_topology_cpy


class _TestInstanceNUMACell(object):

    def test_siblings(self):
        # default thread number of VirtualCPUTopology is one, one thread means
        # no thread and no sibling
        inst_cell = objects.InstanceNUMACell(
            cpuset=set([0, 1, 2]), pcpuset=set())
        self.assertEqual([], inst_cell.siblings)
        inst_cell = objects.InstanceNUMACell(
            cpuset=set([0, 1, 2]), pcpuset=set([4, 5, 6]))
        self.assertEqual([], inst_cell.siblings)

        # 'threads=0' means no sibling
        topo = objects.VirtCPUTopology(sockets=1, cores=3, threads=0)
        inst_cell = objects.InstanceNUMACell(
            cpuset=set([0, 1, 2]), pcpuset=set(), cpu_topology=topo)
        self.assertEqual([], inst_cell.siblings)
        inst_cell = objects.InstanceNUMACell(
            cpuset=set(), pcpuset=set([0, 1, 2]), cpu_topology=topo)
        self.assertEqual([], inst_cell.siblings)

        # One thread actually means no threads
        topo = objects.VirtCPUTopology(sockets=1, cores=3, threads=1)
        inst_cell = objects.InstanceNUMACell(
            cpuset=set([0, 1, 2]), pcpuset=set(), cpu_topology=topo)
        self.assertEqual([], inst_cell.siblings)
        inst_cell = objects.InstanceNUMACell(
            cpuset=set(), pcpuset=set([0, 1, 2]), cpu_topology=topo)
        self.assertEqual([], inst_cell.siblings)

        # 2 threads per virtual core, and numa node has only one type CPU
        # pinned and un-pinned.
        topo = objects.VirtCPUTopology(sockets=1, cores=2, threads=2)
        inst_cell = objects.InstanceNUMACell(
            cpuset=set([0, 1, 2, 3]), pcpuset=set(), cpu_topology=topo)
        self.assertEqual([set([0, 1]), set([2, 3])], inst_cell.siblings)
        inst_cell = objects.InstanceNUMACell(
            cpuset=set(), pcpuset=set([0, 1, 2, 3]), cpu_topology=topo)
        self.assertEqual([set([0, 1]), set([2, 3])], inst_cell.siblings)

        # 4 threads per virtual core, numa node has only one type CPU
        topo = objects.VirtCPUTopology(sockets=1, cores=1, threads=4)
        inst_cell = objects.InstanceNUMACell(
            cpuset=set([0, 1, 2, 3]), pcpuset=set(), cpu_topology=topo)
        self.assertEqual([set([0, 1, 2, 3])], inst_cell.siblings)
        inst_cell = objects.InstanceNUMACell(
            cpuset=set(), pcpuset=set([0, 1, 2, 3]), cpu_topology=topo)
        self.assertEqual([set([0, 1, 2, 3])], inst_cell.siblings)

        # 2 threads per virtual core, numa node with two type CPUs, the pinned
        # and un-pinned
        topo = objects.VirtCPUTopology(sockets=1, cores=2, threads=2)
        inst_cell = objects.InstanceNUMACell(
            cpuset=set([0, 1]), pcpuset=set([2, 3]), cpu_topology=topo)
        self.assertEqual([set([0, 1]), set([2, 3])], inst_cell.siblings)

    def test_pin(self):
        inst_cell = objects.InstanceNUMACell(
            cpuset=set([4, 5]), pcpuset=set([0, 1, 2, 3]), cpu_pinning=None)

        # Only vCPU in the 'pcpuset' list is eligible for pinning
        inst_cell.pin(0, 14)
        self.assertEqual({0: 14}, inst_cell.cpu_pinning)

        # vCPU 12 is not a CPU of this cell, drop silently
        inst_cell.pin(12, 14)
        self.assertEqual({0: 14}, inst_cell.cpu_pinning)

        # vCPU in the 'cpuset' which is for floating CPUs, drop silently
        inst_cell.pin(4, 15)
        self.assertEqual({0: 14}, inst_cell.cpu_pinning)

        # Another vCPU appeared in 'pcpuset', ready for pinning
        inst_cell.pin(1, 16)
        self.assertEqual({0: 14, 1: 16}, inst_cell.cpu_pinning)

    def test_pin_vcpus(self):
        inst_cell = objects.InstanceNUMACell(
            cpuset=set([4, 5]), pcpuset=set([0, 1, 2, 3]), cpu_pinning=None)

        # 'pcpuset' is the vCPU list for pinning
        inst_cell.pin_vcpus((0, 14), (1, 15), (2, 16), (3, 17))
        self.assertEqual({0: 14, 1: 15, 2: 16, 3: 17}, inst_cell.cpu_pinning)

        # vCPU out of 'pcpuset' list will not be added to the CPU pinning list
        inst_cell.pin_vcpus((0, 14), (1, 15), (2, 16), (3, 17), (4, 18))
        self.assertEqual({0: 14, 1: 15, 2: 16, 3: 17}, inst_cell.cpu_pinning)

        # vCPU not belonging to this cell will be dropped silently
        inst_cell.pin_vcpus((0, 14), (1, 15), (2, 16), (3, 17), (10, 18))
        self.assertEqual({0: 14, 1: 15, 2: 16, 3: 17}, inst_cell.cpu_pinning)

    def test_cpu_pinning(self):
        topo_obj = get_fake_obj_numa_topology(self.context)
        self.assertEqual(set(), topo_obj.cpu_pinning)

        topo_obj.cells[0].pin_vcpus((1, 10), (2, 11))
        self.assertEqual(set([10, 11]), topo_obj.cpu_pinning)

        topo_obj.cells[1].pin_vcpus((3, 0), (4, 1))
        self.assertEqual(set([0, 1, 10, 11]), topo_obj.cpu_pinning)

    def test_cpuset_reserved(self):
        topo_obj = get_fake_obj_numa_topology(self.context)
        self.assertEqual(set([5, 6, 7, 8]), topo_obj.cpuset_reserved)

    def test_clear_host_pinning(self):
        topo_obj = get_fake_obj_numa_topology(self.context)
        topo_obj.cells[0].pin_vcpus((1, 10), (2, 11))
        topo_obj.cells[0].id = 3
        topo_obj.cells[1].pin_vcpus((3, 0), (4, 1))
        topo_obj.cells[1].id = 0

        topo_obj.clear_host_pinning()

        self.assertEqual({}, topo_obj.cells[0].cpu_pinning)
        self.assertEqual(-1, topo_obj.cells[0].id)
        self.assertEqual({}, topo_obj.cells[1].cpu_pinning)
        self.assertEqual(-1, topo_obj.cells[1].id)

    def test_emulator_threads_policy(self):
        topo_obj = get_fake_obj_numa_topology(self.context)
        self.assertFalse(topo_obj.emulator_threads_isolated)
        topo_obj.emulator_threads_policy = (
            fields.CPUEmulatorThreadsPolicy.ISOLATE)
        self.assertTrue(topo_obj.emulator_threads_isolated)

    def test_obj_make_compatible(self):
        topo_obj = objects.InstanceNUMACell(
            cpuset=set(), pcpuset=set([0, 1]),
            cpuset_reserved=set([1, 2]),
            cpu_policy=fields.CPUAllocationPolicy.MIXED,
        )
        versions = ovo_base.obj_tree_get_versions('InstanceNUMACell')
        data = lambda x: x['nova_object.data']

        primitive = data(topo_obj.obj_to_primitive(
            target_version='1.6', version_manifest=versions))
        self.assertEqual(
            fields.CPUAllocationPolicy.MIXED, primitive['cpu_policy'])

        self.assertRaises(
            exception.ObjectActionError,
            topo_obj.obj_to_primitive,
            target_version='1.5', version_manifest=versions)

        # set this to something compatible with <  1.6 so we can keep testing
        topo_obj.cpu_policy = fields.CPUAllocationPolicy.DEDICATED
        primitive = data(topo_obj.obj_to_primitive(
            target_version='1.5', version_manifest=versions))
        self.assertIn('pcpuset', primitive)

        primitive = data(topo_obj.obj_to_primitive(
            target_version='1.4', version_manifest=versions))
        self.assertNotIn('pcpuset', primitive)
        self.assertEqual(set([0, 1]), set(primitive['cpuset']))
        self.assertIn('cpuset_reserved', primitive)

        primitive = data(topo_obj.obj_to_primitive(
            target_version='1.3', version_manifest=versions))
        self.assertNotIn('cpuset_reserved', primitive)


class TestInstanceNUMACell(
    test_objects._LocalTest, _TestInstanceNUMACell,
):
    pass


class TestInstanceNUMACellRemote(
    test_objects._RemoteTest, _TestInstanceNUMACell,
):
    pass


class _TestInstanceNUMATopology(object):

    @mock.patch('nova.db.main.api.instance_extra_update_by_uuid')
    def test_create(self, mock_update):
        topo_obj = get_fake_obj_numa_topology(self.context)
        topo_obj.instance_uuid = fake_db_topology['instance_uuid']
        topo_obj.create()
        self.assertEqual(1, len(mock_update.call_args_list))

    def _test_get_by_instance_uuid(self):
        numa_topology = objects.InstanceNUMATopology.get_by_instance_uuid(
            self.context, fake_db_topology['instance_uuid'])
        self.assertEqual(fake_db_topology['instance_uuid'],
                         numa_topology.instance_uuid)
        for obj_cell, topo_cell in zip(
                numa_topology.cells, fake_obj_numa_topology['cells']):
            self.assertIsInstance(obj_cell, objects.InstanceNUMACell)
            self.assertEqual(topo_cell.id, obj_cell.id)
            self.assertEqual(topo_cell.cpuset, obj_cell.cpuset)
            self.assertEqual(topo_cell.memory, obj_cell.memory)
            self.assertEqual(topo_cell.pagesize, obj_cell.pagesize)

    @mock.patch('nova.db.main.api.instance_extra_get_by_instance_uuid')
    def test_get_by_instance_uuid(self, mock_get):
        mock_get.return_value = fake_db_topology
        self._test_get_by_instance_uuid()

    @mock.patch('nova.db.main.api.instance_extra_get_by_instance_uuid')
    def test_get_by_instance_uuid_missing(self, mock_get):
        mock_get.return_value = None
        self.assertRaises(
            exception.NumaTopologyNotFound,
            objects.InstanceNUMATopology.get_by_instance_uuid,
            self.context, 'fake_uuid')

    def test_cpu_policy(self):
        cpu_policy = fields.CPUAllocationPolicy.SHARED
        topology = objects.InstanceNUMATopology(
            instance_uuid=fake_instance_uuid,
            cells=[
                objects.InstanceNUMACell(
                    cpuset=set([0, 1, 2, 3]),
                    pcpuset=set(),
                    cpu_pinning=None,
                    cpu_policy=cpu_policy,
                ),
                objects.InstanceNUMACell(
                    cpuset=set([4, 5, 6, 7]),
                    pcpuset=set(),
                    cpu_pinning=None,
                    cpu_policy=cpu_policy,
                ),
            ],
        )

        self.assertEqual(cpu_policy, topology.cpu_policy)

    def test_cpu_policy__error(self):
        """Ensure we raise an error if cells have different values."""
        topology = objects.InstanceNUMATopology(
            instance_uuid=fake_instance_uuid,
            cells=[
                objects.InstanceNUMACell(
                    cpuset=set([0, 1, 2, 3]),
                    pcpuset=set(),
                    cpu_pinning=None,
                    cpu_policy=None,
                ),
                objects.InstanceNUMACell(
                    cpuset=set([4, 5, 6, 7]),
                    pcpuset=set(),
                    cpu_pinning=None,
                    cpu_policy=fields.CPUAllocationPolicy.SHARED
                ),
            ],
        )

        with testtools.ExpectedException(exception.InternalError):
            topology.cpu_policy

    def test_cpuset_reserved(self):
        topology = objects.InstanceNUMATopology(
            instance_uuid=fake_instance_uuid,
            cells=[
                objects.InstanceNUMACell(
                    id=0, cpuset=set([1, 2]), pcpuset=set(), memory=512,
                    pagesize=2048, cpuset_reserved=set([3, 7])),
                objects.InstanceNUMACell(
                    id=1, cpuset=set([3, 4]), pcpuset=set(), memory=512,
                    pagesize=2048, cpuset_reserved=set([9, 12]))
            ])
        self.assertEqual(set([3, 7]), topology.cells[0].cpuset_reserved)
        self.assertEqual(set([9, 12]), topology.cells[1].cpuset_reserved)

    def test_obj_make_compatible_numa(self):
        topo_obj = objects.InstanceNUMATopology(
            emulator_threads_policy=(
                fields.CPUEmulatorThreadsPolicy.ISOLATE))
        versions = ovo_base.obj_tree_get_versions('InstanceNUMATopology')
        primitive = topo_obj.obj_to_primitive(target_version='1.2',
                                              version_manifest=versions)
        self.assertNotIn(
            'emulator_threads_policy', primitive['nova_object.data'])

        topo_obj = objects.InstanceNUMATopology.obj_from_primitive(primitive)
        self.assertFalse(topo_obj.emulator_threads_isolated)

    def test_obj_from_db_obj(self):
        """Test of creating 'InstanceNUMATopology' OVO object from the
        database primitives, which has an old version 'InstanceNUMACell'
        primitives.

        Prior to version 1.5, 'InstanceNUMACell' saves the instance CPUs in the
        'cpuset' field, for both the pinned CPUs of a dedicated and the
        un-pinned CPUs of a shared instances, after version 1.5, any pinned
        CPUs of dedicated instance are moved to 'pcpuset'. this test verifies
        the CPU movement for instance with a 'dedicated' allocation policy.
        """
        fake_topo_obj_w_cell_v1_4 = objects.InstanceNUMATopology(
            instance_uuid=fake_instance_uuid,
            cells=[
                objects.InstanceNUMACell(
                    id=0, cpuset=set([1, 2]), pcpuset=set(), memory=512,
                    pagesize=2048),
                objects.InstanceNUMACell(
                    id=1, cpuset=set([3, 4]), pcpuset=set(), memory=512,
                    pagesize=2048),
            ])

        fake_topo_obj = copy.deepcopy(fake_topo_obj_w_cell_v1_4)
        for cell in fake_topo_obj.cells:
            cell.cpu_policy = objects.fields.CPUAllocationPolicy.DEDICATED
            cell.VERSION = '1.4'

        numa_topology = objects.InstanceNUMATopology.obj_from_db_obj(
            self.context, fake_instance_uuid, fake_topo_obj._to_json())

        for obj_cell, topo_cell in zip(
                numa_topology.cells,
                fake_topo_obj_w_cell_v1_4['cells']):
            self.assertEqual(set(), obj_cell.cpuset)
            self.assertEqual(topo_cell.cpuset, obj_cell.pcpuset)

    def test_obj_from_db_obj_no_pinning(self):
        """Test of creating 'InstanceNUMATopology' OVO object from the
        database primitives, which has an old version 'InstanceNUMACell'
        primitives.

        Prior to version 1.5, 'InstanceNUMACell' saves the instance CPUs in
        the 'cpuset' field, for both the pinned CPUs of a dedicated and the
        un-pinned CPUs of a shared instances, after version 1.5, any pinned
        CPUs of dedicated instance are moved to 'pcpuset'. this test verifies
        the CPU movement for instance with a 'dedicated' allocation policy.

        This test is for the case where the instance has no pinned CPUs but
        the instance has a numa topology such as when hugepages are used.
        See bug: https://bugs.launchpad.net/nova/+bug/2080556 for more details.
        """
        fake_topo_obj_w_cell_v1_4 = objects.InstanceNUMATopology(
            instance_uuid=fake_instance_uuid,
            cells=[
                objects.InstanceNUMACell(
                    id=0, cpuset=set([1, 2]), memory=512,
                    pagesize=2048),
                objects.InstanceNUMACell(
                    id=1, cpuset=set([3, 4]), memory=512,
                    pagesize=2048),
            ])

        fake_topo_obj = copy.deepcopy(fake_topo_obj_w_cell_v1_4)
        for cell in fake_topo_obj.cells:
            cell.cpu_policy = objects.fields.CPUAllocationPolicy.SHARED
            cell.VERSION = '1.4'

        numa_topology = objects.InstanceNUMATopology.obj_from_db_obj(
            self.context, fake_instance_uuid, fake_topo_obj._to_json())

        for obj_cell, topo_cell in zip(
                numa_topology.cells,
                fake_topo_obj_w_cell_v1_4['cells']):
            self.assertEqual(topo_cell.cpuset, obj_cell.cpuset)
            self.assertEqual(set(), obj_cell.pcpuset)

    def test__migrate_legacy_dedicated_instance_cpuset_dedicate_1_4(self):
        # Create a topology with a cell on latest version. Would be nice
        # to create one the old 1.4 cell version directly but that is only
        # possible indirectly as done below.
        topo = objects.InstanceNUMATopology(
            instance_uuid=fake_instance_uuid,
            cells=[
                objects.InstanceNUMACell(id=0, cpuset=set(), pcpuset={0, 1}),
            ])
        topo.cells[0].cpu_policy = objects.fields.CPUAllocationPolicy.DEDICATED

        # Use the builtin backlevelling logic to pull it back to old cell
        # version
        topo_with_cell_1_4 = topo.obj_to_primitive(
            target_version='1.3', version_manifest={'InstanceNUMACell': '1.4'})

        # Just check that the backlevelling works, and we have a cell with
        # version and data on 1.4 level
        cell_1_4_primitive = topo_with_cell_1_4['nova_object.data']['cells'][0]
        self.assertEqual('1.4', cell_1_4_primitive['nova_object.version'])
        self.assertEqual(
            (0, 1), cell_1_4_primitive['nova_object.data']['cpuset'])
        self.assertNotIn('pcpuset', cell_1_4_primitive['nova_object.data'])

        # Now simulate that such old data is loaded from the DB and migrated
        # from 1.4 to 1.6 by the data migration
        topo_loaded = objects.InstanceNUMATopology.obj_from_db_obj(
            self.context, fake_instance_uuid,
            jsonutils.dumps(topo_with_cell_1_4))

        # In place data migration happens during load, cpuset data is moved to
        # pcpuset
        self.assertEqual(set(), topo_loaded.cells[0].cpuset)
        self.assertEqual({0, 1}, topo_loaded.cells[0].pcpuset)
        # and the version is updated to 1.6
        self.assertEqual('1.6', topo_loaded.cells[0].VERSION)

    def test__migrate_legacy_dedicated_instance_cpuset_shared_1_4(self):
        # Create a topology with a cell on latest version. Would be nice
        # to create one the old 1.4 cell version directly but that is only
        # possible indirectly as done below.
        topo = objects.InstanceNUMATopology(
            instance_uuid=fake_instance_uuid,
            cells=[
                objects.InstanceNUMACell(id=0, cpuset={0, 1}, pcpuset=set()),
            ])
        topo.cells[0].cpu_policy = objects.fields.CPUAllocationPolicy.SHARED

        # Use the builtin backlevelling logic to pull it back to old cell
        # version
        topo_with_cell_1_4 = topo.obj_to_primitive(
            target_version='1.3', version_manifest={'InstanceNUMACell': '1.4'})

        # Just check that the backlevelling works, and we have a cell with
        # version and data on 1.4 level
        cell_1_4_primitive = topo_with_cell_1_4['nova_object.data']['cells'][0]
        self.assertEqual('1.4', cell_1_4_primitive['nova_object.version'])
        self.assertEqual(
            (0, 1), cell_1_4_primitive['nova_object.data']['cpuset'])
        self.assertNotIn('pcpuset', cell_1_4_primitive['nova_object.data'])

        # Now simulate that such old data is loaded from the DB and migrated
        # from 1.4 to 1.6 by the data migration
        topo_loaded = objects.InstanceNUMATopology.obj_from_db_obj(
            self.context, fake_instance_uuid,
            jsonutils.dumps(topo_with_cell_1_4))

        # In place data migration did not move the data as the cpu_policy is
        # shared
        self.assertEqual({0, 1}, topo_loaded.cells[0].cpuset)
        self.assertEqual(set(), topo_loaded.cells[0].pcpuset)
        # but the version is updated to 1.6
        self.assertEqual('1.6', topo_loaded.cells[0].VERSION)

    def test__migrate_legacy_dedicated_instance_cpuset_dedicated_half_migrated(
        self
    ):
        # Before the fix for https://bugs.launchpad.net/nova/+bug/2097360
        # landed Nova only half migrated the InstanceNUMACell from 1.4 to 1.6
        # by doing the data move but not updating the version string in the DB.
        # This test case ensures that if such migration happened before the fix
        # was deployed then Nova fixed the DB content on the next load as well.

        # Create a topology with a cell on latest version. Would be nice
        # to create one the old 1.4 cell version directly but that is only
        # possible indirectly as done below.
        topo = objects.InstanceNUMATopology(
            instance_uuid=fake_instance_uuid,
            cells=[
                objects.InstanceNUMACell(id=0, cpuset=set(), pcpuset={0, 1}),
            ])
        topo.cells[0].cpu_policy = objects.fields.CPUAllocationPolicy.DEDICATED

        # simulate the half done migration by pulling back the version string
        # in the primitive form to 1.4 while keeping the data on 1.6 format.
        topo_primitive = topo.obj_to_primitive()
        cell_primitive = topo_primitive['nova_object.data']['cells'][0]
        self.assertEqual('1.6', cell_primitive['nova_object.version'])
        cell_primitive['nova_object.version'] = '1.4'

        topo_loaded = objects.InstanceNUMATopology.obj_from_db_obj(
            self.context, fake_instance_uuid, jsonutils.dumps(topo_primitive))

        # the data did not change
        self.assertEqual(set(), topo_loaded.cells[0].cpuset)
        self.assertEqual({0, 1}, topo_loaded.cells[0].pcpuset)
        # but the version is updated to 1.6
        self.assertEqual('1.6', topo_loaded.cells[0].VERSION)

    def test__migrate_legacy_dedicated_instance_cpuset_mixed_1_4(self):
        # Before the fix for https://bugs.launchpad.net/nova/+bug/2097360
        # landed Nova only half migrated the InstanceNUMACell from 1.4 to 1.6
        # by doing the data move but not updating the version string in the DB.
        # After this the instance is resized to flavor with mixed cpu_policy
        # then there is a good chance that the DB has version string 1.4 but
        # the data is on 1.6 format with the cpu_policy set to mixed.
        # This test case ensures that if such situation happened before the fix
        # was deployed then Nova fixed the DB content on the next load as well.

        # Create a topology with a cell on latest version. Would be nice
        # to create one the old 1.4 cell version directly but that is only
        # possible indirectly as done below.
        topo = objects.InstanceNUMATopology(
            instance_uuid=fake_instance_uuid,
            cells=[
                objects.InstanceNUMACell(id=0, cpuset={0, 1}, pcpuset={2, 3}),
            ])
        topo.cells[0].cpu_policy = objects.fields.CPUAllocationPolicy.MIXED

        # simulate the half done migration by pulling back the version string
        # in the primitive form to 1.4 while keeping the data on 1.6 format.
        topo_primitive = topo.obj_to_primitive()
        cell_primitive = topo_primitive['nova_object.data']['cells'][0]
        self.assertEqual('1.6', cell_primitive['nova_object.version'])
        cell_primitive['nova_object.version'] = '1.4'

        topo_loaded = objects.InstanceNUMATopology.obj_from_db_obj(
            self.context, fake_instance_uuid, jsonutils.dumps(topo_primitive))

        # the data did not change
        self.assertEqual({0, 1}, topo_loaded.cells[0].cpuset)
        self.assertEqual({2, 3}, topo_loaded.cells[0].pcpuset)
        self.assertEqual(
            objects.fields.CPUAllocationPolicy.MIXED,
            topo_loaded.cells[0].cpu_policy)
        # but the version is updated to 1.6
        self.assertEqual('1.6', topo_loaded.cells[0].VERSION)


class TestInstanceNUMATopology(
    test_objects._LocalTest, _TestInstanceNUMATopology,
):
    pass


class TestInstanceNUMATopologyRemote(
    test_objects._RemoteTest, _TestInstanceNUMATopology,
):
    pass
